/*
The dictionary class reads the text file
 */


class Dictionary {
  Node tree;

  int minSugLen = 0; //words that extend the beginning only by n chars won't be suggested

  String path; //file path 



  Dictionary(String path) {
    tree = new Node(' ');
    suggestions = new ArrayList<RankedWord>();

    String[] content = loadStrings(path);
    if (content != null) {
      parseTree(content[0]);
    }

    this.path = path;
  }



  class Node {
    char letter;
    int rank; // wie hoch wird dieser Node bewertet (wie oft wurde das Wort schon verwendet)

    Node children[] = {};

    Node(char c) {
      letter = c;
    }
    Node(char c, int v) {
      letter = c; // char for this node
      rank = v;   // usage value for this word
    }

    void addChild(char c) {
      Node copy[] = new Node[children.length + 1];
      System.arraycopy(children, 0, copy, 0, children.length);
      children = copy;
      children[children.length - 1] = new Node(c);
    }
    void addChild(Node c) {
      Node copy[] = new Node[children.length + 1];
      System.arraycopy(children, 0, copy, 0, children.length);
      children = copy;
      children[children.length - 1] = c;
    }
  }





  //adding a full word will create all nodes that are missing and if the word is already existing it increments the rank
  void addWord(String word) {
    Node index = tree;
    for (int i = 0; i < word.length(); i++) {

      int position = searchNode(index, word.charAt(i));

      if (position < 0) {
        index.addChild(word.charAt(i));
        position = index.children.length - 1;
      }
      index = index.children[position];
    }
    index.rank = min(index.rank + 1, 1000);
  }




  // search a beginning of a word in the tree
  // in the process all suggestions will be stored in the suggestions list of the dictionary
  void searchWord(String beginning) {
    if (beginning.length() > 1 ) {
      Node index = tree;
      suggestions.clear();
      for (int i = 0; i < beginning.length(); i++) {

        int position = searchNode(index, beginning.charAt(i));

        if (position < 0) 
          return;

        index = index.children[position];
      }

      //now search for suggestions via recursion
      searchingRecursion(index, "");

      //post edit th suggestion list: sort
      Collections.sort(suggestions, new Comparator<RankedWord>() {
        @Override
          public int compare(RankedWord s1, RankedWord s2) {
          return s2.rank - s1.rank;
        }
      }
      );
      String beginning2 = beginning.substring(0, beginning.length()-1);
      for (int i = 0; i < suggestions.size(); i++) {
        suggestions.get(i).word = beginning2 + suggestions.get(i).word;
      }
    } else {
      suggestions.clear();
    }
  }


  //StringList suggestions;
  ArrayList<RankedWord> suggestions;

  // the rank is needed after searching is almost completed so the words can be suggested in the right order. 
  // therefore we have this little class
  class RankedWord {
    String word;
    int rank;
    RankedWord(String w, int r) {
      word = w;
      rank = r;
    }
  }




  //sugeestion search recursion
  void searchingRecursion(Node n, String carr) {
    if (n.rank > 0 && carr.length() > minSugLen) 
      suggestions.add(new RankedWord(carr + n.letter, n.rank));

    for (int i = 0; i < n.children.length; i++) 
      searchingRecursion(n.children[i], carr + n.letter);
  }

  // search a node for a child with the specified letter, returns the index of that child in the array
  int searchNode(Node n, char c) {
    for (int i = 0; i < n.children.length; i++) { 
      if (n.children[i].letter == c) {
        return i;
      }
    }
    return -1;
  }


  // list up all words of the dictionary
  void suggestAllWords() {
    suggestions.clear();
    searchingRecursion(tree, "");
  }


  //former storing option: every word is stored list-like in a text file
  @Deprecated
    void saveDictionary() {
    suggestAllWords();
    String allWords[] = new String[suggestions.size() + 1];
    allWords[0] = ""; //this entry is needed because saveStrings seems to add an invisible char at beginning of file
    println(allWords.length);

    for (int i = 0; i < suggestions.size(); i++) {
      allWords[i+1] = trim(suggestions.get(i).word);
    }

    saveStrings("dict.txt", allWords);
  }


  // parse the tree from the text file
  void parseTree(String text) {
    this.text = text;
    this.index = 0;
    parseNode(tree);
    text = "";
  }


  String text; //content of listfile, used for reading AND writing (when saving the file)
  int index = 0; //when parsing this int stores the position of the char in the string from textfile that is curently processed 
  //(needed because the recursion doesn't pass this index


  //recursion for parsing all nodes
  void parseNode(Node n) {
    char letter = 0;
    int rank = 0;


    while (true) {

      char c = 0;

      if (index < text.length()) 
        c = text.charAt(index);
      else  break;


      index++;

      if (c >= 0 && c <= 9) { //digit

        rank = rank * 10 + c;
      } else if (c == ' ') {

        Node newNode = new Node(letter, rank);
        n.addChild(newNode);

        parseNode(newNode);

        rank = 0;
        letter = 0;
      } else if (c == '/') { //closing slash
        /*if (letter != 0) {
         Node newNode = new Node(letter, rank);
         n.addChild(newNode);
         //index--; //count back becaues now we advanced a bit too far
         //parseNode(newNode);
         }*/

        return;
      } else { //char

        if (letter == 0) {
          letter = c;
        } else {
          Node newNode = new Node(letter, rank);
          n.addChild(newNode);

          index--; //count back becaues now we advanced a bit too far

          parseNode(newNode);

          rank = 0;
          letter = 0;
        }
      }
    }
  }


  //save dictionary content to the same file it came from
  void saveData() {
    text = "";
    nodesToString(tree);
    //println(text);

    String a[] = {trim(text)};
    saveStrings(path, a);
  }

  //recursion
  void nodesToString(Node n) {
    int r = n.rank;
    String rank = r > 0 ? "" : ("" + (char)0);

    //convert rank to string
    while (r > 0) {
      rank = (char)((r % 10)) + rank;
      r /= 10;
    }
    text += n.letter;
    if (n.rank > 0)
      text += rank + " ";


    for (int i = 0; i < n.children.length; i++) 
      nodesToString(n.children[i]);

    text += "/";
  }

  //
}
