
package pGUI.core;

import java.lang.reflect.Field;

import pGUI.classes.*;

public class Animation {

	Object target;
	Field field;
	String attributeName;
	String fieldType; // "int", "double", "float" etc.

	int animationType; // enable i.e. color animations
	static final protected int NUMBER = 0;
	static final protected int COLOR = 1;

	// number animation steps (frames) the animations needs to complete
	int numberOfSteps;
	int counter = 0;

	double currentValue;

	// for numbers
	double valueStart;
	double valueEnd;

	// for colors
	int a1, r1, g1, b1;
	int a2, r2, g2, b2;

	public Animation(String attribute, Object target, int aimedValue, double time) {

		this.target = target;

		if (time < 100000) { // prevent very long or infinite animation, especially when steps are negative
			// println("animate", attribute);
			try {
				this.attributeName = attribute;
				field = getField(target.getClass(), attribute);
				field.setAccessible(true);

				fieldType = field.getType().toString();

				// get initial value
				currentValue = field.getDouble(target);

				// calculate needed number of frames to complete animation
				numberOfSteps = (int) (((Control) target).papplet.frameRate * time / 1000);

				if (attribute.contains("Color")) {
					animationType = COLOR;
					a1 = ((int) currentValue >> 24) & 0xff;
					r1 = ((int) currentValue >> 16) & 0xff;
					g1 = ((int) currentValue >> 8) & 0xff;
					b1 = ((int) currentValue) & 0xff;

					a2 = (aimedValue >> 24) & 0xff;
					r2 = (aimedValue >> 16) & 0xff;
					g2 = (aimedValue >> 8) & 0xff;
					b2 = (aimedValue) & 0xff;
					//
				} else {
					animationType = NUMBER;
					this.valueStart = currentValue;
					this.valueEnd = aimedValue;
				}
			} catch (NoSuchFieldException e) {
				e.printStackTrace();
			} catch (IllegalAccessException ia) {
				ia.printStackTrace();
			}
		}
	}

	protected boolean animate() {

		// check if there's still work to do
		if (counter <= numberOfSteps) {
			switch (animationType) {
			case NUMBER:
				currentValue = valueStart + (valueEnd - valueStart) / (float) numberOfSteps * counter;
				break;

			case COLOR:
				double ac = (a1 + (a2 - a1) / (float) numberOfSteps * counter);
				double rc = (r1 + (r2 - r1) / (float) numberOfSteps * counter);
				double gc = (g1 + (g2 - g1) / (float) numberOfSteps * counter);
				double bc = (b1 + (b2 - b1) / (float) numberOfSteps * counter);

				currentValue = Color.colors((int) rc, (int) gc, (int) bc, (int) ac);
				// println(rc, gc, bc, ac);
				break;
			}
		} else {
			return false; // end animation with false, which clears it off animation queue (in Frame)
		}

		counter++;

		// set value
		try {
			switch (fieldType) {
			case "int":
				field.setInt(target, (int) currentValue);
				break;
			case "float":
				field.setInt(target, (int) currentValue);
				break;
			case "double":
				field.setInt(target, (int) currentValue);
				break;
			case "long":
				field.setInt(target, (int) currentValue);
				break;
			}
		} catch (IllegalAccessException ia) {
			ia.printStackTrace();
		}
		((Control) target).update();
		return true;
	}

	// get even protected fields
	private Field getField(Class<?> clazz, String fieldName) throws NoSuchFieldException {
		try {
			return clazz.getDeclaredField(fieldName);
		} catch (NoSuchFieldException e) {
			Class<?> superClass = clazz.getSuperclass();
			if (superClass == null) {
				throw e;
			} else {
				return getField(superClass, fieldName);
			}
		}
	}

	// comparing function, returns true if it's the same attribute on the same
	// object
	public boolean compare(Animation other) {
		if (this.attributeName.equals(other.attributeName) && this.target == other.target)
			return true;
		else
			return false;
	}
}