package pGUI.core;



import processing.core.*;
import processing.event.*;

import java.lang.reflect.Method;

import pGUI.classes.*;


public class Checkbox extends Control {

  protected int UncheckedBackgroundColor;
  protected int CheckedBackgroundColor;
  protected int SwitchForegroundColor;
  protected int Size; //size of switch, equals height of the checkbox 

  protected boolean checked = false;

  public Checkbox(PApplet pa) {
    super(pa);

    Size = 20;
    Height = Size;

    UncheckedBackgroundColor = Color.colors(160);
    CheckedBackgroundColor = Color.colors(46, 116, 122);
    SwitchForegroundColor = Color.colors(250);

    setBackgroundColor(0);

    //need to create new array because parents constructor doesn't know about the new numberOfListeners
    numberOfListeners = 6; //added one listener: itemChangedListener
    registeredListeners = new boolean[numberOfListeners];
    registeredMethods = new Method[numberOfListeners];
    registeredMethodWithArgs = new boolean[numberOfListeners];

    addMouseListener();
  }

  protected void render() {
    pg = papplet.createGraphics(Width, Height);
    pg.beginDraw();

    if (VisualBackgroundColor != 0) {
      pg.background(VisualBackgroundColor);
    }

    pg.fill(checked ? CheckedBackgroundColor : UncheckedBackgroundColor);
    pg.stroke(BorderColor);
    pg.rect(1, 1, Size-2, Size-2, 2);


    //draw check mark

    if (checked) {
      pg.strokeWeight(2);
      pg.stroke(SwitchForegroundColor);
      pg.line(Size / 4, Size/2, Size/2.5f, 3 * Size / 4);
      pg.line(Size/2.5f, 3 * Size / 4, 3 * Size / 4, Size/4);
    }
    pg.noStroke();

    pg.textSize(FontSize);
    pg.fill(ForegroundColor);
    pg.textAlign(37, 3);
    pg.text(Text, Size + PaddingLeft + Size / 4, Height/2);

    if (!Enabled) {
      pg.fill(150, 150);
      pg.rect(1, 1, Size-2, Size-2, 2);
    }


    pg.endDraw();
  }




  /*
   *  SETTER
   */

  public void setUncheckedBackgroundColor(int clr) {
    UncheckedBackgroundColor = clr;
  }
  public void setCheckedBackgroundColor(int clr) {
    CheckedBackgroundColor = clr;
  }
  public void setSwitchForegroundColor(int clr) {
    SwitchForegroundColor = clr;
  }
  public void setSize(int size) {
    Size = size;
    autosize();
  }

  @Override 
    protected void autosize() {
    pg = papplet.createGraphics(1, 1);
    pg.beginDraw();
    pg.textSize(FontSize);

    Width = (int)(1.75 * Size + 10 + pg.textWidth(Text) + PaddingLeft + PaddingRight);
    Height = (int)PApplet.max(Size, FontSize + pg.textDescent(), 1);
  }
  
  
  public void setChecked(boolean state) {
    if (Enabled) {
      checked = state;
      handleRegisteredEventMethod(5, null);
      update();
    }
  }




  public int getUncheckedBackgroundColor() {
    return UncheckedBackgroundColor;
  } 
  public int getCheckedBackgroundColor() {
    return CheckedBackgroundColor;
  } 
  public int getSwitchForegroundColor() {
    return SwitchForegroundColor;
  } 
  public int getSize() {
    return Size;
  } 
  public boolean isChecked() {
    return checked;
  } 




  /*
   * EVENTS
   */
  public void addCheckedChangedListener(String methodName) {
    registerEventMethod(5, methodName);
  }
  public void removeCheckedChangedListener() {
    deregisterEventMethod(5);
  }

  protected void press(MouseEvent e) {
    if (Enabled) {
      checked = !checked;
      handleRegisteredEventMethod(5, null);
      update();
    }
  }
}
