
package pGUI.core; //<>// //<>// //<>//

import processing.core.*;
import processing.event.*;

import java.util.ArrayList;
import java.awt.EventQueue;
import java.util.Collections;
import java.util.Comparator;

import pGUI.classes.*;

public class Container extends Control {

	// child items
	protected ArrayList<Control> Content;
	protected boolean AutoSize = false;

	public Container(PApplet pa) {
		super(pa);
		Width = 100;
		Height = 100;
		Content = new ArrayList<Control>();
	}

	/*
	 * DRAWING AND RENDERING
	 */

	@Override
	protected void render() {
		// if AutoSize is on, first get minimal dimensions
		if (AutoSize) {
			setAutoSize();
		}
		pg = papplet.createGraphics(Width, Height);
		pg.beginDraw();
		drawDefaultBackground();

		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);
			if (c.Visible) {

				// do this before render, so containers know their bounds when rendering their
				// items
				// set bounds of child so it has absolute values for listener processing
				c.Bounds.X0 = Math.max(this.Bounds.X0 + c.X, this.Bounds.X0);
				c.Bounds.Y0 = Math.max(this.Bounds.Y0 + c.Y, this.Bounds.Y0);
				// crop overflow
				c.Bounds.X = Math.min(this.Bounds.X0 + c.X + c.Width, this.Bounds.X);
				c.Bounds.Y = Math.min(this.Bounds.Y0 + c.Y + c.Height, this.Bounds.Y);

				if (c.changedVisuals) {

					c.changedVisuals = false; // do this before render, so it can use animations by calling update again
					c.render();
				}
				pg.image(c.getGraphics(), c.X, c.Y);
			}
		}
		pg.endDraw();
	}

	// not only setting this Frame-property but also all it's children
	@Override
	protected void setParentFrame(Frame f) {
		this.frame = f;
		for (int i = 0; i < Content.size(); i++) {
			Content.get(i).setParentFrame(f);
		}
	}

	public void add(Control c) {
		Content.add(c);
		c.parent = this;
		if (c.frame != this.frame) {
			c.frame = this.frame;
		}
		update();
	}

	public void add(Control... controls) {
		for (Control c : controls) {
			add(c);
		}
		update();
	}

	protected void sortContent() {
		Collections.sort(Content, new Comparator<Control>() {
			@Override
			public int compare(Control c1, Control c2) {
				return c1.Z - c2.Z;
			}
		});
	}

	public void clear() {
		Content.clear();
		update();
	}

	public void remove(int index) {
		Content.remove(index);
		update();
	}

	public void remove(Control c) {
		Content.remove(c);
		update();
	}

	public Control[] getItems() {
		Control c[] = new Control[Content.size()];
		for (int i = 0; i < Content.size(); i++) {
			c[i] = Content.get(i);
		}
		return c;
	}

	// resize container so it fits its content
	public void setAutoSize() {
		int mWidth = 1;
		int mHeight = 1;
		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);
			if (c.Visible) {
				if (c.X + c.Width > mWidth) {
					mWidth = c.X + c.Width;
				}
				if (c.Y + c.Height > mHeight) {
					mHeight = c.Y + c.Height;
				}
			}
		}
		Width = mWidth;
		Height = mHeight;
	}

	@Override
	public void mouseEvent(MouseEvent e) {
		if (Visible) {

			// reverse iteration direction (as to drawing) so topmost elements will
			// get the chance to choke the listener for objects below
			for (int i = Content.size() - 1; i >= 0; i--) {
				if (!frame.chokeMouseListeners) {
					if (Content.get(i).registeredMouseListener)
						Content.get(i).mouseEvent(e);
					i = Math.min(i, Content.size()); // it's possible that contents have changed meanwhile
				} else {
					break;
				}
			}
			if (!frame.chokeMouseListeners)
				super.mouseEvent(e);
		}
	}
}