package pGUI.core;

import processing.core.*;
import processing.event.*;

import pGUI.classes.*;

public class HScrollContainer extends HFlowContainer {

	// Width the scrollContainer would have considering all its content
	// (fullScrollWidth can actually be smaller that Width!)
	protected int fullScrollWidth;

	// Scroll position (only horizontal), starts at 0
	protected int ScrollPosition = 0;

	// speed at which container will be scrolled, can be set externally
	protected int ScrollSpeed = 10;

	public HScrollContainer(PApplet pa) {
		super(pa);
		addMouseListener();
	}

	@Override
	protected void render() {
		// if AutoSize is on, first get minimal dimensions
		if (AutoSize) {
			setAutoSize();
		}

		pg = papplet.createGraphics(Width, Height);
		pg.beginDraw();
		drawDefaultBackground();

		int usedSpace = 0;

		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);

			if (c.Visible) {
				c.Bounds.X0 = this.Bounds.X0 + usedSpace + c.MarginLeft + PaddingLeft - ScrollPosition;
				c.Bounds.Y0 = this.Bounds.Y0 + MarginTop + PaddingTop;

				c.Bounds.X = Math.min(c.Bounds.X0 + c.Width, this.Bounds.X);
				c.Bounds.Y = Math.min(c.Bounds.Y0 + c.Height, this.Bounds.Y);

				// constrain after computing X,Y so no data will be lost by constraining
				c.Bounds.X0 = Math.max(c.Bounds.X0, this.Bounds.X0);
				c.Bounds.Y0 = Math.max(c.Bounds.Y0, this.Bounds.Y0);

				if (c.changedVisuals) {
					c.changedVisuals = false;
					c.render();
				}

				// subtract scrollPositionX from x-drawing-value
				pg.image(c.getGraphics(), usedSpace + c.MarginLeft + PaddingLeft - ScrollPosition,
						c.MarginTop + PaddingTop);
				usedSpace += (c.Width + c.MarginLeft + c.MarginRight);
			}
		}

		// get total virtual width of all elements rowed up
		fullScrollWidth = PaddingLeft;
		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);

			if (c.Visible) {
				fullScrollWidth += c.MarginRight + c.Width + c.MarginLeft;
			}
		}

		// Draw view-only scrollbar:
		float scrollbarWidth = (float) Width / fullScrollWidth * Width;

		if (scrollbarWidth < Width) { // don't display scroll-bar when there's nothign to scroll

			float scrollBarX = PApplet.constrain(
					ScrollPosition * (Width - scrollbarWidth) / (float) (fullScrollWidth - Width), 1,
					Width - scrollbarWidth - 2);
			pg.fill(150);
			pg.noStroke();
			pg.rect(scrollBarX, Height + 1, scrollbarWidth, 3, 15);
		}

		pg.endDraw();
	}

	public void setScrollPositionX(float i) {
		ScrollPosition = PApplet.constrain((int) i, 0, PApplet.max(0, fullScrollWidth - Width));
		update();
	}

	/*
	 * SETTER
	 */

	public void setScrollSpeed(int s) {
		ScrollSpeed = s;
	}

	public int getScrollSpeed() {
		return ScrollSpeed;
	}

	/*
	 * MOUSE EVENTS
	 */

	@Override
	protected void mouseWheel(MouseEvent e) {
		setScrollPositionX(ScrollPosition + e.getCount() * ScrollSpeed);
	}
}