
package pGUI.core;

import processing.core.*;
import processing.event.*;

import java.util.ArrayList;
import java.lang.reflect.Method;

import pGUI.classes.*;

public class ListView extends VScrollContainer {

  protected int ItemBackgroundColor;
  protected int SelectionColor;
  protected int SelectionHoverColor;
  protected boolean MultiSelect = false;

  protected Control selectedItem;
  protected int selectionIndex = -1;

  protected ArrayList<Control> selectedItems;

  protected int selectedItemOriginalBGColor;
  protected int selectedItemOriginalHoverColor;


  public ListView(PApplet pa) {
    super(pa);
    selectedItems = new ArrayList<Control>();

    //need to create new array because parents constructor doesn't know about the new numberOfListeners
    numberOfListeners = 6; //added one listener: itemChangedListener
    registeredListeners = new boolean[numberOfListeners];
    registeredMethods = new Method[numberOfListeners];
    registeredMethodWithArgs = new boolean[numberOfListeners];

    ItemBackgroundColor = -1; //white
    setBackgroundColor(-3618616); //light grey
    FontSize = 20;
    TextAlign = 37;
    Width = 100;
    Height = 150;
    setSelectionColor(-12171706);
  }





  protected void itemSelected(Control c) {
    if (!controlDown && !shiftDown) { //only deselect selected when control not pressed
      if (!MultiSelect || selectedItem != null) {
        deselect(selectionIndex);
      }
      if (MultiSelect) {
        for (int i = 0; i < Content.size(); i++) {
          if (selectedItems.contains(Content.get(i))) {
            deselect(i);
            selectedItems.remove(Content.get(i));
          }
        }
      }
    }

    //get index
    int index = -1;
    for (int i = 0; i < Content.size(); i++) {
      if (c == Content.get(i)) index = i;
    }

    if (MultiSelect) {
      if (shiftDown) {
        for (int i = selectionIndex < index ? selectionIndex : index; i <= (selectionIndex > index ? selectionIndex : index); i++) {
          select(i);
          selectedItems.add(Content.get(i));
        }
      } else {
        if (selectedItems.contains(c)) {
          deselect(index);
          selectedItems.remove(c);
        } else {
          select(index);
          selectedItems.add(c);
        }
      }
    } else {
      select(index);
    }
    //println(selectedItems);

    /*
    for (int i = 0; i < Content.size(); i++) {
     if (shiftDown) {
     if (i > selectionIndex+1) {
     selectedItems.add(Content.get(i-1));
     select(i-1);
     }
     }
     if (c == Content.get(i)) {
     if (MultiSelect) {
     if (selectedItems.contains(c)) {
     deselect(i);
     selectedItems.remove(c);
     } else {
     select(i);
     selectedItems.add(c);
     }
     } else {
     select(i);
     }
     println(selectedItems);
     break;
     }
     }*/
  }

  public void add(String...newItems) {
    for (String newItem : newItems) {
      add(newItem);
    }
  }


  public void add(String newItem) {
    ListItem newListItem = new ListItem(papplet);

    /* Some properties will be taken over from the listview to make adding simple listitems by string easier. 
     Of course this works only for properties that listview itself doesn't use */
    newListItem.Text = newItem;
    newListItem.FontSize = FontSize;
    newListItem.TextAlign = TextAlign;

    newListItem.setBackgroundColor(ItemBackgroundColor);
    newListItem.setForegroundColor(ForegroundColor);
    this.add(newListItem);
  }


  protected void select(int index) {
    selectedItem = Content.get(index);
    selectionIndex = index;

    selectedItemOriginalBGColor = selectedItem.BackgroundColor;
    selectedItemOriginalHoverColor = selectedItem.HoverColor;

    selectedItem.VisualBackgroundColor = SelectionColor;
    selectedItem.BackgroundColor = SelectionColor;
    selectedItem.HoverColor = SelectionHoverColor;
    selectedItem.update();
    
  
    handleRegisteredEventMethod(5, null);
  }


  protected void deselect(int index) {
    if (index >= 0 && index < Content.size()) {
      Content.get(index).VisualBackgroundColor = selectedItemOriginalBGColor;
      Content.get(index).BackgroundColor = selectedItemOriginalBGColor;
      Content.get(index).HoverColor = selectedItemOriginalHoverColor;
      Content.get(index).update();
    }
  }

  public void scrollToItem(int index) {
    if (index >= 0 && index < Content.size()) {
      float y = 0;
      for (int i = 0; i < index; i++) {
        y += Content.get(i).MarginTop + Content.get(i).Height + Content.get(i).MarginBottom;
      }
      if (ScrollPosition > y  ) {
        setScrollPositionY(y);
      } else if (ScrollPosition + Height < y + Content.get(index).Height) {
        setScrollPositionY(y - Height + Content.get(index).Height + Content.get(index).MarginTop + Content.get(index).MarginBottom);
      }
    }
  }







  /*
   * SETTER
   */


  public void setItemBackgroundColor(int clr) {
    ItemBackgroundColor = clr;
    update();
  }
  public void setSelectionColor(int clr) {
    SelectionColor = clr;
    int r = (int)papplet.red(clr);
    int g = (int)papplet.green(clr);
    int b = (int)papplet.blue(clr);
    int a = (int)papplet.alpha(clr);

    if (papplet.brightness(clr) > 40) { //darken color for HoverColor and PressedColor when color is bright enough
      SelectionHoverColor = Color.colors(r-20, g-20, b-20);
    } else {//lighten color for HoverColor and PressedColor when color too dark
      SelectionHoverColor = Color.colors(r+20, g+20, b+20);
    }
    update();
  }
  public void setSelectedItem(int index) {
    if (MultiSelect) {
      if (index == selectionIndex) {
      }
      if (selectedItem != null) {
        deselect(selectionIndex);
      }
      if (index >= 0 && index < Content.size()) {
        select(index);
        scrollToItem(index);
      }
    } else { //if single select
      if (selectedItem != null) {
        deselect(selectionIndex);
      }
      if (index >= 0 && index < Content.size()) {
        select(index);
        scrollToItem(index);
      }
    }
    update();
  }
  public void setMultiSelect(boolean state) {
    MultiSelect = state;
  }

  /*
   * GETTER
   */


  public int getItemBackgroundColor() {
    return ItemBackgroundColor;
  }
  public int getSelectionColor() {
    return SelectionColor;
  }
  public Control getSelectedItem() {
    return selectedItem;
  }
  public int getSelectionIndex() {
    return selectionIndex;
  }
  public ArrayList<Control> getSelectedItems() {
    return selectedItems;
  }
  public boolean getMultiSelect() {
    return MultiSelect;
  }

  /*
   * EVENTS
   */

  public void addItemChangedListener(String methodName) {
    registerEventMethod(5, methodName);
  }
  public void removeItemChangedListener() {
    deregisterEventMethod(5);
  }

  @Override
    protected void drag(MouseEvent e) {
  }
  @Override
    protected void press(MouseEvent e) {
    this.Focus();
  }
  protected boolean controlDown = false;
  protected boolean shiftDown = false;
  @Override
    public void onKeyPress(KeyEvent e) {
    if (Enabled) {
      switch(e.getKeyCode()) {
      case PApplet.CONTROL:
        controlDown = true;
        break;
      case PApplet.SHIFT:
        shiftDown = true;
        break;
      case PApplet.DOWN:
        if (selectionIndex < Content.size()-1) {
          deselect(selectionIndex);
          select(selectionIndex+1);
          scrollToItem(selectionIndex);
        }
        break;
      case PApplet.UP:
        if (selectionIndex > 0) {
          deselect(selectionIndex);
          select(selectionIndex-1);
          scrollToItem(selectionIndex);
        }
        break;
      }
    }
  }
  @Override
    public void onKeyRelease(KeyEvent e) {
    if (Enabled) {
      switch(e.getKeyCode()) {
      case PApplet.CONTROL:
        controlDown = false;
        break;
      case PApplet.SHIFT:
        shiftDown = false;
        break;
      }
    }
  }
}
