
package pGUI.core;

import processing.core.*;
import processing.event.*;

import java.lang.reflect.Method;

import pGUI.classes.*;

public class MenuStrip extends VScrollContainer {

	protected boolean open = false;

	protected int HeadHeight = 25;
	protected int DropDownWidth = 1;
	protected int DropDownBackgroundColor = 240;

	protected Control selectedItem;
	protected int selectionIndex;

	public MenuStrip(PApplet pa) {
		super(pa);
		Text = "SPINNER";
		Width = 90;
		Height = 30;
		BorderWidth = 0;
		FontSize = 15;
		setPadding(0, 10, 4, 10);

		BackgroundColor = 0;
		VisualBackgroundColor = 0;
		HoverColor = 1342177280; // just darken menucontainer backgroundcolor
		PressedColor = 1677721600;

		// need to create new array because parents constructor doesn't know about the
		// new numberOfListeners
		numberOfListeners = 6; // added one listener: itemChangedListener
		registeredListeners = new boolean[numberOfListeners];
		registeredMethods = new Method[numberOfListeners];
		registeredMethodWithArgs = new boolean[numberOfListeners];
	}

	@Override
	protected void render() {
		if (!open) {
			renderClosed();
		} else {
			renderOpen();
		}
	}

	protected void renderClosed() {
		pg = papplet.createGraphics(1, 1);
		pg.beginDraw();
		pg.textSize(FontSize);
		Width = (int) (pg.textWidth(Text) + PaddingLeft + PaddingRight);
		Height = HeadHeight;
		pg = papplet.createGraphics(Width, Height);
		pg.beginDraw();

		// only draw head
		drawDefaultBackground();
		drawDefaultText();

		pg.endDraw();
		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);
			c.Bounds.X0 = 0;
			c.Bounds.Y0 = 0;
			c.Bounds.X = 0;
			c.Bounds.Y = 0;
		}
		Bounds.Y = this.Bounds.Y0 + Height;
	}

	protected void renderOpen() {
		// first(!) get needed dropdown width
		int DropDownWidth = 0;
		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);
			if (c.Visible) {
				int minControlWidth = (int) (c.PaddingLeft + c.PaddingRight + pg.textWidth(c.Text)
						+ pg.textWidth(((MenuItem) c).shortcutToText()) + FontSize * 2);
				DropDownWidth = minControlWidth > DropDownWidth ? minControlWidth : DropDownWidth;
			}
		}

		int WidthBackup = Width;
		// do this for short time, so the children know the width
		Width = DropDownWidth;

		// get total virtual height of all elements lined up
		fullScrollHeight = 0;
		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);
			if (c.Visible) {
				// do rendering here, so fullscrollheight can be calculated properly in first
				// attempt
				if (c.changedVisuals) {
					c.changedVisuals = false;
					c.render();
				}
				fullScrollHeight += c.MarginTop + c.Height + c.MarginBottom;

				pg.textSize(c.FontSize);
			}
		}
		Width = WidthBackup;

		Height = HeadHeight + fullScrollHeight;
		pg = papplet.createGraphics(Math.max(DropDownWidth, Width) + 5, Height + 5);
		pg.beginDraw();

		drawShadow(DropDownWidth, Height - HeadHeight, 5);

		// draw dropdownbackground
		pg.stroke(-5592406); // color(170)
		pg.strokeWeight(1);
		pg.fill(DropDownBackgroundColor);
		pg.rect(0, HeadHeight, DropDownWidth, Height - HeadHeight);

		// draw head
		Height = HeadHeight;
		// only draw head
		drawDefaultBackground();
		drawDefaultText();
		Height = HeadHeight + fullScrollHeight;

		// draw the little vertical line
		pg.strokeWeight(1);
		pg.stroke(220);
		pg.line(22, HeadHeight + 3, 22, Height - 3);
		pg.stroke(255);
		pg.line(23, HeadHeight + 3, 23, Height - 3);

		int usedSpace = HeadHeight;

		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);

			if (c.Visible) {
				c.Bounds.X0 = this.Bounds.X0 + c.MarginLeft + PaddingLeft;
				c.Bounds.Y0 = this.Bounds.Y0 + usedSpace + c.MarginTop + PaddingTop - ScrollPosition;

				c.Bounds.X = (c.Bounds.X0 + c.Width); // dont cut bounds here!!
				c.Bounds.Y = Math.min(c.Bounds.Y0 + c.Height, this.Bounds.Y);

				c.Bounds.X0 = Math.max(c.Bounds.X0, this.Bounds.X0);
				c.Bounds.Y0 = Math.max(c.Bounds.Y0, this.Bounds.Y0);

				// no paddingleft here!
				pg.image(c.getGraphics(), c.MarginLeft, usedSpace + c.MarginTop + PaddingTop - ScrollPosition);
				usedSpace += (c.Height + c.MarginTop + c.MarginBottom);
			}
		}
		// extend bounds to dropdown width
		Bounds.X = DropDownWidth + Bounds.X0;

		pg.endDraw();
	}

	public void add(String newItem) {
		MenuItem newMenuItem = new MenuItem(papplet);
		newMenuItem.Text = newItem;
		this.add(newMenuItem);
	}

	public void add(String... newItems) {
		for (String newItem : newItems) {
			MenuItem newMenuItem = new MenuItem(papplet);
			newMenuItem.Text = newItem;
			this.add(newMenuItem);
		}
	}

	/*
	 * GETTER & SETTER
	 */

	protected void drawShadow(int w, int h, int offset) {
		pg.noFill();
		int cl[] = { 115, 85, 41, 15, 5 };
		for (int i = 0; i < 5; i++) {
			pg.stroke(Color.colors(0, cl[i]));
			pg.rect((offset - 1) * 2 - i, HeadHeight + (offset - 1) * 2 - i, w - 2 * (4 - i), h - 2 * (4 - i));
		}
	}

	@Override
	public void setBackgroundColor(int clr) {
		setBackgroundColorsAutomatically(clr);
	}

	public void openDropDown() {
		((MenuContainer) parent).openMenu(this);

		VisualBackgroundColor = PressedColor;
		open = true;
		update();
	}

	public void closeDropDown() {
		((MenuContainer) parent).openMenu = null;
		VisualBackgroundColor = BackgroundColor;
		open = false;
		update();
	}

	public boolean isDroppedDown() {
		return open;
	}

	public Control getSelectedItem() {
		return selectedItem;
	}

	public int getSelectionIndex() {
		return selectionIndex;
	}

	protected void itemSelected(Control c) {
		int index = -1;
		for (int i = 0; i < Content.size(); i++) {
			if (c == Content.get(i))
				index = i;
		}
		// just double-check
		if (index >= 0) {
			selectedItem = c;
			selectionIndex = index;
			handleRegisteredEventMethod(5, null);
		}
		closeDropDown();
	}

	/*
	 * EVENTS
	 */
	public void addItemChangedListener(String methodName) {
		registerEventMethod(5, methodName);
	}

	public void removeItemChangedListener() {
		deregisterEventMethod(5);
	}

	@Override
	protected void press(MouseEvent e) {
		if (!open)
			openDropDown();
		else
			closeDropDown();
		update();
	}

	@Override
	protected void enter(MouseEvent e) { // when clicked hovering is sufficient for changing the dropdown
		VisualBackgroundColor = HoverColor;

		if (((MenuContainer) parent).openMenu != null) {
			openDropDown();
		}

		update();
	}

	@Override
	protected void exit(MouseEvent e) {
		VisualBackgroundColor = BackgroundColor;
		update();
	}

	@Override
	protected void release(MouseEvent e) {
		update();
	}

	// @Override
	public void mouseEvent(MouseEvent e) {
		super.mouseEvent(e);
		int x = e.getX();
		int y = e.getY();

		// close dropdown when clicked elsewhere
		if (open && (x < Bounds.X0 || x > Bounds.X || y > Bounds.Y)) { // if not over element
			if (e.getAction() == MouseEvent.PRESS || e.getAction() == MouseEvent.WHEEL) {
				closeDropDown();
			}
		}
		// always choke mouselisteners, so hover events won't occur for underlying
		// objects
		if (x > Bounds.X0 && x < Bounds.X && y > Bounds.Y0 + HeadHeight && y < Bounds.Y) { // if not over element
			frame.chokeMouseListeners = true;
		}
	}
}
