package pGUI.core; //<>// //<>// //<>// //<>// //<>//



import processing.core.*;
import processing.event.*;
import processing.data.StringList;
import processing.data.IntList;

import java.lang.reflect.Method;

//clipboard
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.StringSelection;
import java.io.*;
import java.awt.datatransfer.DataFlavor;


import pGUI.classes.*;






public class MultilineTextbox extends VScrollContainer {

  /*
   *  STYLE
   */


  protected StringList lines;        //text in lines
  protected IntList lineBreaks;      //marks the automatic and entered line-breaks


  //style
  protected int CursorColor = 70;
  protected int SelectionColor = Color.colors(51, 153, 255);
  protected String Hint = "";



  protected int cursorPosition = 0;
  protected boolean clickSetsCursor = true;

  //not to set by the user
  protected int lineHeight;
  protected boolean autoScroll; //autoscroll to cursor when text changed, dont set it

  protected int cursorTime;
  protected boolean displayCurs = true;
  protected boolean initialized = false;

  // protected  //add textChangedListener



  public MultilineTextbox(PApplet pa) {
    super(pa);
    lines = new StringList();
    lineBreaks = new IntList();
    ForegroundColor = 0;
    lineHeight = FontSize + 10;
    setPadding(3);
    setSize(100, 70);
    BorderWidth = 0;
    setBackgroundColor(230);
    overridesParentsShortcuts = true;
    Cursor = PApplet.TEXT;
    //addMouseListener(); already done by VScrollContainer

    //need to create new array because parents constructor doesn't know about the new numberOfListeners
    numberOfListeners = 6; //text-changed listener
    registeredListeners = new boolean[numberOfListeners];
    registeredMethods = new Method[numberOfListeners];
    registeredMethodWithArgs = new boolean[numberOfListeners];

    //cursor animation
    papplet.registerMethod("pre", this);
  }


  /*
   * Graphics
   */


  @Override
    protected void render() {
    if (!initialized) {
      cursorTime = papplet.millis();
      initialized = true;
    }

    pg = papplet.createGraphics(Width, Height); //add some extra space for the scroll bar
    pg.beginDraw();
    pg.textSize(FontSize);
    pg.textAlign(PApplet.LEFT, PApplet.TOP);
    drawDefaultBackground();


    if (BorderWidth == 0) {
      //draw 3D-Border
      pg.strokeWeight(1);
      pg.stroke(70);
      pg.line(0, 0, Width, 0);
      pg.line(0, 0, 0, Height);
    }

    // turn 'Text'-String to lines-list. Also compute breaks
    boxedText(Text, X, Y, Width); //actually takes the most time with long text

    //do this before drawing cursor
    fullScrollHeight = lines.size() * lineHeight + PaddingTop + PaddingBottom;


    //DRAW CURSOR, before drawing text!
    if (this.Focused && displayCurs) {
      drawCursor();
    }

    //DRAW SELECTION
    if (selectionStart < selectionEnd && Focused) {
      if (selectionStart <= Text.length() && selectionEnd <= Text.length()) {

        getLineToCursor(selectionStart);
        getLineToCursor(selectionStart);

        pg.fill(SelectionColor);
        pg.noStroke();

        for (int i = getLineToCursor(selectionStart); i <= getLineToCursor(selectionEnd); i++) {
          int start = Math.max(selectionStart, lineBreaks.get(i));
          int end = Math.min(selectionEnd, lineBreaks.get(i+1)-1);

          int selectionX = (int)pg.textWidth(lines.get(i).substring(0, start - lineBreaks.get(i)));

          int selectionWidth = (int)pg.textWidth(lines.get(i).substring(start - lineBreaks.get(i), end - lineBreaks.get(i)));
          pg.rect(PaddingLeft + selectionX, i * lineHeight + PaddingTop - ScrollPosition, selectionWidth, FontSize + pg.textDescent());
        }
      }
    }

    //DRAW TEXT
    pg.fill(ForegroundColor);
    if (!Text.equals("")) {
      for (int i = 0; i < lines.size(); i++) {
        pg.text(lines.get(i), PaddingLeft, i * lineHeight + PaddingTop - ScrollPosition);
      }
    } else { //draw hint
      pg.fill(120);
      pg.text(Hint, PaddingLeft, PaddingTop);
    }

    //DRAW SCROLLBAR
    float scrollbarHeight = (float)Height / fullScrollHeight * Height;
    if (scrollbarHeight < Height) { //don't display scroll-bar when there's nothing to scroll

      float scrollBarY = PApplet.constrain(ScrollPosition * (Height - scrollbarHeight) / (float)(fullScrollHeight - Height), 1, Height - scrollbarHeight - 2);
      pg.fill(150);
      pg.noStroke();
      pg.rect(Width - 4, scrollBarY, 3, scrollbarHeight, 15);
    }

    pg.endDraw();
  }









  /* turn text to stringlist and compute all line break positions */

  protected void boxedText(String str, int x0, int y0, int w) {
    lines = new StringList();
    lineBreaks = new IntList();

    lineHeight = FontSize * 3 / 2;

    float spaceWidth = pg.textWidth(" ");
    float usedLineSpace = 0;
    String currentLine = "";
    int countChars = 0;

    lineBreaks.append(0);

    w = w - PaddingLeft - PaddingRight;

    String paragraphs[] = PApplet.split(str, '\n');

    for (int j = 0; j < paragraphs.length; j++) {
      String words[] = PApplet.split(paragraphs[j], ' ');

      for (int i = 0; i < words.length; i++) {
        float wordWidth = pg.textWidth(words[i]);
        if (wordWidth < w  ) { 
          if (usedLineSpace + wordWidth < w) { //no new line

            currentLine += words[i] + " ";
            usedLineSpace += spaceWidth + wordWidth;
            countChars += 1 + words[i].length();
          } else { //break line

            countChars += 1;
            lineBreaks.append(countChars-1);     //store position of break
            lines.append(currentLine);        //store previous line

            currentLine = words[i] + " ";      //new word is first word of currentLine

            usedLineSpace = wordWidth + spaceWidth;
            countChars += words[i].length();
          }
        } else { //the only word in this line is too long

          //first store currentLine if it isn't empty
          if (!currentLine.equals("")) {
            lineBreaks.append(countChars);     //store position of break
            lines.append(currentLine);        //store previous line

            currentLine = "";      //new word is first word of currentLine

            usedLineSpace = 0;
          }

          for (int k = words[i].length(); k > 0; k--) { //go back step by step and see at what point it fits

            wordWidth -= pg.textWidth(words[i].substring(k-1, k));

            if (wordWidth < w) { //found substring that fits line
              currentLine += words[i].substring(0, k-1);
              countChars += words[i].substring(0, k-1).length();

              lineBreaks.append(countChars);      //store position of break
              lines.append(currentLine);          //store previous line

              words[i] = words[i].substring(k-1); //set this word to remainder
              i--;                                //.. and go back a step in the iteration

              currentLine = "";                   //reset new line
              usedLineSpace = 0;

              break;
            }
          }
        }
      }
      currentLine += "\n"; 
      lineBreaks.append(countChars);   
      lines.append(currentLine); 

      currentLine = "";
      usedLineSpace = 0;
    }
    lineBreaks.append(countChars);
  }




  /*
   * CURSOR
   */


  //cursor blink animation
  public void pre() {
    if (this.Focused) {
      int t = papplet.millis();
      if (t - cursorTime > 500) {
        displayCurs = !displayCurs;
        cursorTime = t;
        update();
      }
    }
  }





  protected void drawCursor() {
    int index = getLineToCursor(cursorPosition);
    int cursorHeight = FontSize;
    if (index >= 0 && cursorPosition > 0) {
      int start = lineBreaks.get(index);
      int stop = cursorPosition;

      String a = Text.substring(start, stop);

      a = lines.get(index).substring(0, cursorPosition-start);
      float wordWidth = pg.textWidth(a) + (float)0.5;

      int cursorY = (index) * lineHeight; //position of upper left corner of cursor relative to first character of text

      //autoscroll
      if (autoScroll) {
        if (cursorY - ScrollPosition + cursorHeight > Height - PaddingBottom - PaddingTop) {   //cursor left visible box at the bottom
          setScrollPositionY(cursorY - Height + FontSize + PaddingTop + PaddingBottom);
        } else if (cursorY < ScrollPosition + PaddingTop) {          //cursor left visible box at the top
          setScrollPositionY(cursorY);
        }
        autoScroll = false;
      }

      //draw cursor
      pg.stroke(CursorColor);
      pg.line(PaddingLeft + wordWidth, cursorY - ScrollPosition + PaddingTop, wordWidth + PaddingLeft, cursorY + cursorHeight - ScrollPosition + PaddingTop);
    } else {
      pg.stroke(CursorColor);
      pg.line(PaddingLeft, PaddingTop, PaddingLeft, cursorHeight + PaddingTop);
    }
  }



  //take a cursor position and get the line the cursor should be in
  protected int getLineToCursor(int cursor) { 
    int cursorLine = 1;

    for (int i = 1; i < lineBreaks.size(); i++) {
      if (cursor < lineBreaks.get(i)) {
        cursorLine = i - 1;
        break;
      }
    }
    return cursorLine;
  }





  /*
   * EDIT TEXT
   */

  protected void append(char c) {
    cursorPosition += 1;
    Text = Text.substring(0, cursorPosition-1) + c + Text.substring(cursorPosition-1);

    textChanged();
  }

  protected void append(String s) {
    Text = Text.substring(0, cursorPosition) + s + Text.substring(cursorPosition);
    cursorPosition += s.length();

    textChanged();
  }


  protected void backspace() {
    if (Text.length() > 0 && cursorPosition >= 1) {

      Text = Text.substring(0, cursorPosition-1) + Text.substring(cursorPosition);
      cursorPosition--;
      textChanged();
    }
  }

  protected void deleteKey() {
    if (Text.length() > cursorPosition) {

      Text = Text.substring(0, cursorPosition) + Text.substring(cursorPosition+1);
      textChanged();
    }
  }


  protected void deleteRange(int start, int end) {
    start = start >=0 ? (start <= Text.length() ? start : Text.length()) : 0;
    end = end >=0 ? (end <= Text.length() ? end : Text.length()) : 0;
    if (start > end) {
      int temp = start;
      start = end;
      end = temp;
    }
    textChanged();
    Text = Text.substring(0, start) + Text.substring(end);
  }

  protected void textChanged() {
    handleRegisteredEventMethod(5, null);
    cursorChanged();
  }
  protected void cursorChanged() {
    cursorTime = papplet.millis();
    displayCurs = true;
    selectionStart = cursorPosition;
    selectionEnd = cursorPosition;
    autoScroll = true;
    update();
  }

  protected void moveCursorBy(int i) {
    cursorPosition = PApplet.constrain(cursorPosition + i, 0, Text.length());
    cursorChanged();
  }
  protected void moveCursorTo(int i) {
    cursorPosition = PApplet.constrain(i, 0, Text.length());
    cursorChanged();
  }





  protected void moveCursorVertically(int direction) { //negative down, positive up
    int oldCursorLine = getLineToCursor(cursorPosition); //textline, in which cursor has been so far

    if (oldCursorLine-direction >= 0 && oldCursorLine-direction < lines.size()) {
      String oldSubstring = Text.substring(lineBreaks.get(oldCursorLine), cursorPosition); //get line up to cursor where cursor has been so far
      float widthPreviousLineToCursor = pg.textWidth(oldSubstring);                        //and the width in pixels

      String newLineText = lines.get(oldCursorLine-direction);                             //text of the aimed line

      float wide = 0;        //variable for counting the width in next step

      float letterWidth = 0;

      for (int i = 0; i < newLineText.length()-1; i++) {                                   //find out at what letter of the new line the approximately same width in pixels is reached
        String a = newLineText.substring(i, i+1);
        letterWidth = pg.textWidth(a);
        wide += letterWidth;

        if (wide-letterWidth/2 >= widthPreviousLineToCursor) {        //set decision point to the center of the letter
          moveCursorTo(lineBreaks.get(oldCursorLine-direction) + i);
          break;
        }
      }

      if (wide-letterWidth/2 < widthPreviousLineToCursor) {           //in case previous line has been longer than aimed line
        if (direction > 0) {
          moveCursorTo(lineBreaks.get(oldCursorLine)-1);              //for up moving
        } else {    
          moveCursorTo(lineBreaks.get(oldCursorLine-2*direction)-1);  //for down moving
        }
      }
    }
  }

  protected int findNextWord() {
    //in first phase search for next space, in second search for first letter
    int phase = 0;
    for (int i = cursorPosition; i < Text.length(); i++) {
      if (phase == 0) {
        if (Text.charAt(i) == ' ') {
          phase = 1;
        }
      } else {
        if (Text.charAt(i) != ' ' && Text.charAt(i) != '\n')
          return i;
      }
    }
    return Text.length() - 1;
  }
  protected int findPreviousWord() {
    //in first phase search for next space, in second search for first letter
    int phase = 0;
    for (int i = cursorPosition-1; i > 0; i--) {
      if (phase == 0) {
        if (Text.charAt(i) == ' ') {
          phase = 1;
        }
      } else {
        if (Text.charAt(i) != ' ' && Text.charAt(i) != '\n')
          return i+1;
      }
    }
    return 0;
  }




  /*
   * GETTER AND SETTER
   */

  public void setCursorColor(int clr) {
    this.CursorColor = clr;
    update();
  }
  public void setCursorPosition(int p) {
    this.cursorPosition = p>=0 ? (p<=Text.length() ? p : Text.length()) : 0;
    selectionEnd = cursorPosition;
    selectionStart = cursorPosition;
    update();
  }
  public void setSelectionColor(int clr) {
    SelectionColor = clr;
  }
  public void setHint(String h) {
    Hint = h;
  }
  public void setClickSetsCursor(boolean state) {
    clickSetsCursor = state;
  }

  public int getCursorColor() {
    return CursorColor;
  }
  public int getCursorPosition() {
    return cursorPosition;
  }
  public int getSelectionColor() {
    return SelectionColor;
  } 
  public String getHint() {
    return Hint;
  }
  public boolean getClickSetsCursor() {
    return clickSetsCursor;
  }






  /*
   * EVENTS
   */
  public void addTextChangedListener(String methodName) { //Beware! At the moment this also reacts to setting the cursor
    registerEventMethod(5, methodName);
  }
  public void removeTextChangedListener() {
    deregisterEventMethod(5);
  }

  @Override 
    protected void enter(MouseEvent e) {
    papplet.cursor(Cursor);
    update();
  }
  @Override
    protected void exit(MouseEvent e) {
    papplet.cursor(0);
    update();
  }

  int selectionInitial = 0;
  int selectionStart, selectionEnd;

  @Override
    public void press(MouseEvent e) {
    int mY = e.getY();
    int mX = e.getX();

    //set cursor by clicking
    if (clickSetsCursor) {
      setCursorByClick(mX, mY);

      selectionInitial = cursorPosition;
      selectionStart = cursorPosition;
      selectionEnd = cursorPosition;
    }



    this.Focus();
    cursorTime = papplet.millis();
    displayCurs = true;
    update();
  }

  @Override 
    protected void drag(MouseEvent e) {
    setCursorByClick(e.getX(), e.getY());
    if (cursorPosition > selectionInitial) {
      selectionStart = selectionInitial;
      selectionEnd = cursorPosition;
    } else {
      selectionStart = cursorPosition;
      selectionEnd = selectionInitial;
    }
  }

  protected void  setCursorByClick(int mX, int mY) {
    int clickedPosY = mY - Bounds.Y0 - PaddingTop + ScrollPosition;
    int newCursorLine = clickedPosY / lineHeight;
    String line = "";
    if (lines.size() > newCursorLine) {
      line = lines.get(newCursorLine);
      int clickedPosX = mX - Bounds.X0 - PaddingLeft;  //relative to textbox origin and considering fullScrollWidth

      float wide = 0;
      for (int i = 0; i < line.length(); i++) {
        float letterWidth = pg.textWidth(line.substring(i, i+1));
        wide += letterWidth;
        if (wide-letterWidth/2 > clickedPosX) { //set decision point to the center of the letter
          moveCursorTo(i+lineBreaks.get(newCursorLine));
          break;
        }
      }
      if (wide < clickedPosX) { //in case clicked beyond last letter - set cursor to end
        moveCursorTo(line.length()+lineBreaks.get(newCursorLine)-2);
      }
    } else { //if line number is exceeded, set cursor to end
      moveCursorTo(Text.length());
    }
  }


  boolean ctrl, shft, alt = false;

  @Override 
    public void onKeyRelease(KeyEvent e) {
    switch(e.getKeyCode()) {
    case PApplet.CONTROL:
      ctrl = false;
      break;
    case PApplet.SHIFT:
      shft = false;
      break;
    case PApplet.ALT:
      alt = false;
      break;
    }
  }
  @Override 
    public void onKeyPress(KeyEvent e) {
    if (Enabled) {
      char KEY = e.getKey();
      int Code = e.getKeyCode();


      if (ctrl && !shft && !alt) {
        switch((char)Code) {
        case 'C':
          {
            if (selectionStart < selectionEnd) {
              StringSelection selection = new StringSelection(Text.substring(selectionStart, selectionEnd));
              Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
              clipboard.setContents(selection, selection);
            }
            break;
          }
        case 'V':
          {
            Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
            Transferable contents = clipboard.getContents(null);
            if (contents != null && contents.isDataFlavorSupported(DataFlavor.stringFlavor)) {
              try {
                String content = (String)contents.getTransferData(DataFlavor.stringFlavor);
                if (selectionStart < selectionEnd) { //if there is a selection then replace
                  int selStart = selectionStart; //store this because delete will call cursorChanged which resets selectionStart
                  deleteRange(selectionStart, selectionEnd);
                  moveCursorTo(selStart);
                }
                this.append(content);
              }
              catch(UnsupportedFlavorException ex) {
                System.out.println(ex);
                ex.printStackTrace();
              }
              catch(IOException ex) {
                ex.printStackTrace();
              }
            }
          }
          break;
        case 'A':
          selectionStart = 0;
          selectionEnd = Text.length();
          update();
          break;
        case 'X':
          if (selectionStart < selectionEnd) {
            StringSelection selection = new StringSelection(Text.substring(selectionStart, selectionEnd));
            Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
            clipboard.setContents(selection, selection);

            int selStart = selectionStart; //store this because delete will call cursorChanged which resets selectionStart
            deleteRange(selectionStart, selectionEnd);
            moveCursorTo(selStart);
            selectionStart = 0;
            selectionEnd = 0;
          }


          break;
        }
      }



      if (e.getKey() == PApplet.CODED) {
        switch(Code) {
        case PApplet.LEFT:
          if (ctrl) {
            moveCursorTo(findPreviousWord());
          } else {
            moveCursorBy(-1);
          }
          break;
        case PApplet.RIGHT:  
          if (ctrl) {

            moveCursorTo(findNextWord());
          } else {
            moveCursorBy(1);
          }
          break;
        case PApplet.UP:
          moveCursorVertically(1);
          break;
        case PApplet.DOWN: 
          moveCursorVertically(-1);
          break;
        case PApplet.CONTROL:
          ctrl = true;
          break;
        case PApplet.SHIFT:
          shft = true;
          break;
        case PApplet.ALT:
          alt = true;
          break;
        }
      } else  if (!ctrl) { //not coded
        switch(KEY) {
        case PApplet.BACKSPACE:
          if (selectionStart < selectionEnd) {
            int selStart = selectionStart; //store this because delete will call cursorChanged which resets selectionStart
            deleteRange(selectionStart, selectionEnd);
            moveCursorTo(selStart);
          } else {
            backspace();
          }
          break;
        case PApplet.DELETE:
          if (selectionStart < selectionEnd) {
            int selStart = selectionStart; //store this because delete will call cursorChanged which resets selectionStart
            deleteRange(selectionStart, selectionEnd);
            moveCursorTo(selStart);
          } else {
            deleteKey();
          }
          break;
        default: 
          if (selectionStart < selectionEnd) { //if there is a selection then replace
            int selStart = selectionStart;     //store this because delete will call cursorChanged which resets selectionStart
            deleteRange(selectionStart, selectionEnd);
            moveCursorTo(selStart);
          }
          append(KEY);
        }
      }
    }
  }
}
