package pGUI.core;

import processing.core.*;
import processing.event.*;

import java.lang.reflect.Method;

import pGUI.classes.*;


public class Slider extends Control {

  protected int Value;
  protected int minValue = 0;
  protected int maxValue = 200;
  protected int Size = 15;
  protected int ballsize = 15;


  public Slider(PApplet pa) {
    super(pa);
    Width = 400;
    Value = 100;

    BackgroundColor = -2302756;
    ForegroundColor = -1926085;

    Height = (int)(Size * 1.3);


    numberOfListeners = 6; //added one listener: valueChangedListener
    registeredListeners = new boolean[numberOfListeners];
    registeredMethods = new Method[numberOfListeners];
    registeredMethodWithArgs = new boolean[numberOfListeners];

    addMouseListener();
  }




  @Override
    protected void render() {
    pg = papplet.createGraphics(Width + 2, Height);
    pg.beginDraw();
    pg.noStroke();
    
    int lineHeight = Size /4;
    int ballSize = Size;

    //slider background
    pg.fill(BackgroundColor);
    pg.rect(ballsize / 2, Height/2 - lineHeight/2, Width - ballsize, lineHeight);

    //slider active background
    pg.fill(ForegroundColor);
    pg.rect(ballsize / 2, Height/2 - lineHeight/2, (float)(Width - ballsize) / (maxValue-minValue) * Value, lineHeight);

    //slider ball
    pg.ellipse((float)(Width-ballsize) / Math.abs(maxValue-minValue) * Value + ballsize/2, Height / 2, ballsize, ballsize);

    pg.endDraw();
  }





  /*
   * Getter & Setter
   */


  public void setValue(int value) {
    this.Value = value >= minValue ? (value <= maxValue ? value : maxValue) : minValue;
    update();
  }
  public void setMinValue(int value) {
    minValue = value;
    update();
  }
  public void setMaxValue(int value) {
    maxValue = value;
    update();
  }
  public void setSize(int Size) {
    this.Size = Size;
    Height = (int)(Size * 1.3);
    update();
  }
  public int getValue() {
    return Value;
  }
  public int getMinValue() {
    return minValue;
  }
  public int getMaxValue() {
    return maxValue;
  }
  public int getSize() {
    return Size;
  }






  /*
   * Events
   */

  public void addValueChangedListener(String methodName) {
    registerEventMethod(5, methodName);
  }
  public void removeValueChangedListener() {
    deregisterEventMethod(5);
  }

  protected void setBall(int x) {
    setValue((int)((x - Bounds.X0 - ballsize / 2) * Math.abs(maxValue - minValue) / (float)(Width - ballsize)));
    handleRegisteredEventMethod(5, null);
  }


  @Override
    protected void press(MouseEvent e) {
    setBall(e.getX());
  }
  @Override
    protected void drag(MouseEvent e) {
    setBall(e.getX());
  }
}
