package pGUI.core;

import processing.core.*;
import processing.event.*;

import java.lang.reflect.Method;

//clipboard
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.StringSelection;
import java.io.*;
import java.awt.datatransfer.DataFlavor;

import pGUI.classes.*;

public class Textbox extends HScrollContainer {

  protected int SelectionColor = Color.colors(51, 153, 255);
  protected int CursorColor = 70;
  protected String Hint = "";

  protected int cursorPosition = 0;
  protected int selectionStart = 0;
  protected int selectionEnd = 3;

  protected boolean autoScroll; // autoscroll to cursor when text changed, dont set it

  protected int cursorTime; // variable is used to meausure time between to cursor renders and create the
  // for blink animation
  protected boolean displayCurs = true;
  protected boolean initialized = false;

  public Textbox(PApplet pa) {
    super(pa);

    ForegroundColor = 0;
    BackgroundColor = 220;
    HoverColor = 220;
    VisualBackgroundColor = BackgroundColor;

    FontSize = 20;
    Width = 100;
    Cursor = PApplet.TEXT;
    setPadding(5);
    overridesParentsShortcuts = true;
    autosize();
    // addMouseListener(); already done by HScrollContainer

    // need to create new array because parents constructor doesn't know about the new numberOfListeners
    numberOfListeners = 6; // text-changed listener
    registeredListeners = new boolean[numberOfListeners];
    registeredMethods = new Method[numberOfListeners];
    registeredMethodWithArgs = new boolean[numberOfListeners];

    // cursor animation:
    papplet.registerMethod("pre", this);
  }


  @Override
    protected void render() {
    if (!initialized) {
      cursorTime = papplet.millis();
      initialized = true;
    }

    pg = papplet.createGraphics(Width, Height); // create graphics with new dimensions
    pg.beginDraw();
    drawDefaultBackground();

    if (BorderWidth == 0) {
      // draw 3D-Border
      pg.strokeWeight(1);
      pg.stroke(70);
      pg.line(0, 0, Width, 0);
      pg.line(0, 0, 0, Height);
    }

    // prepare text
    pg.textSize(FontSize);
    pg.textAlign(PApplet.LEFT, PApplet.TOP);

    // do this before drawing cursor!! - needs new fullScrollWidth
    fullScrollWidth = (int) pg.textWidth(Text) + PaddingLeft + PaddingRight;

    // draw cursor
    if (this.Focused && displayCurs) {
      drawCursor();
    }

    // draw selection
    if (selectionStart < selectionEnd && Focused) {
      if (selectionStart <= Text.length() && selectionEnd <= Text.length()) {
        int selectionX = (int) pg.textWidth(Text.substring(0, selectionStart));
        int selectionWidth = (int) pg.textWidth(Text.substring(selectionStart, selectionEnd));
        pg.fill(SelectionColor);
        pg.noStroke();
        pg.rect(PaddingLeft - ScrollPosition + selectionX + FontSize / 15, PaddingTop, 
          selectionWidth + FontSize / 15, Height - PaddingBottom - PaddingTop + pg.textDescent());
      }
    }

    // draw text
    if (!Text.equals("")) {
      pg.fill(ForegroundColor);
      pg.text(Text, PaddingLeft - ScrollPosition, PaddingTop);
    } else {
      pg.fill(120);
      pg.text(Hint, PaddingLeft, PaddingTop);
    }

    // draw scrollbar
    float scrollbarWidth = (float) Width / fullScrollWidth * Width;

    if (scrollbarWidth < Width) { // don't display scroll-bar when there's nothign to scroll
      float scrollBarX = PApplet.constrain(
        ScrollPosition * (Width - scrollbarWidth) / (float) (fullScrollWidth - Width), 1, 
        Width - scrollbarWidth - 2);
      pg.fill(150);
      pg.noStroke();
      pg.rect(scrollBarX, Height - 5, scrollbarWidth, 3, 15);
    }
    pg.endDraw();
  }



  /*
   * Cursor managing
   */

  // cursor blink animation
  public void pre() {
    if (this.Focused) {
      int t = papplet.millis();
      if (t - cursorTime > 500) {
        displayCurs = !displayCurs;
        cursorTime = t;
        update();
      }
    }
  }

  protected void drawCursor() {
    cursorPosition = cursorPosition >= 0 ? (cursorPosition <= Text.length() ? cursorPosition : Text.length()) : 0;
    int start = 0;
    int stop = cursorPosition;// -lineBreaks.get(index-1)+1;
    String a = Text.substring(start, stop);
    float wordWidth = pg.textWidth(a) + 0.5f;
    int cursorHeight = FontSize;

    // do this before drawing the cursor - scrollPositionX has to be set first!!
    if (autoScroll) {
      if (wordWidth - ScrollPosition > Width - PaddingRight - PaddingLeft) { // cursor left visible box at the
        // right
        setScrollPositionX(wordWidth - Width + PaddingRight + PaddingLeft);
      } else if (wordWidth < ScrollPosition + PaddingLeft) { // cursor left visible box at the left
        setScrollPositionX(wordWidth);
      }
      autoScroll = false;
    }
    pg.stroke(CursorColor);
    pg.line(PaddingLeft + wordWidth - ScrollPosition, PaddingTop, wordWidth + PaddingLeft - ScrollPosition, 
      cursorHeight + PaddingTop /*+ pg.textDescent()*/);
  }
  
  
  
  
  
  

  /*
   * EDIT TEXT
   */

  protected void append(char c) {
    if (c != '\n') {
      Text = Text.substring(0, cursorPosition) + c + Text.substring(cursorPosition);
      cursorPosition += 1;
      textChanged();
    }
  }

  protected void append(String s) {
    s = s.replaceAll("\n", "");
    Text = Text.substring(0, cursorPosition) + s + Text.substring(cursorPosition);
    cursorPosition += s.length();
    textChanged();
  }

  protected void backspace() {
    if (Text.length() > 0 && cursorPosition >= 1) {
      Text = Text.substring(0, cursorPosition - 1) + Text.substring(cursorPosition);
      cursorPosition--;
      textChanged();
    }
  }

  protected void deleteKey() {
    if (Text.length() > cursorPosition) {
      Text = Text.substring(0, cursorPosition) + Text.substring(cursorPosition + 1);
      textChanged();
    }
  }

  protected void deleteRange(int start, int end) {
    start = start >= 0 ? (start <= Text.length() ? start : Text.length()) : 0;
    end = end >= 0 ? (end <= Text.length() ? end : Text.length()) : 0;
    if (start > end) {
      int temp = start;
      start = end;
      end = temp;
    }
    textChanged();
    Text = Text.substring(0, start) + Text.substring(end);
  }

  protected void textChanged() {
    handleRegisteredEventMethod(5, null);
    cursorChanged();
  }

  protected void cursorChanged() {
    cursorTime = papplet.millis();
    displayCurs = true;
    selectionStart = cursorPosition;
    selectionEnd = cursorPosition;
    autoScroll = true;
    update();
  }

  protected void moveCursorBy(int i) {
    cursorPosition = PApplet.constrain(cursorPosition + i, 0, Text.length());
    cursorChanged();
  }

  protected void moveCursorTo(int i) {
    cursorPosition = PApplet.constrain(i, 0, Text.length());
    cursorChanged();
  }
  
  

  /*
   * GETTER AND SETTER
   */

  public void setCursorColor(int clr) {
    this.CursorColor = clr;
    update();
  }

  public void setCursorPosition(int p) {
    this.cursorPosition = p >= 0 ? (p <= Text.length() ? p : Text.length()) : 0;
    selectionEnd = cursorPosition;
    selectionStart = cursorPosition;
    update();
  }

  public void setSelectionColor(int clr) {
    SelectionColor = clr;
  }

  public void setHint(String h) {
    this.Hint = h;
  }

  public int getCursorColor() {
    return CursorColor;
  }

  public int getCursorPosition() {
    return cursorPosition;
  }

  public int getSelectionColor() {
    return SelectionColor;
  }

  public String getHint() {
    return Hint;
  }


  @Override
    protected void autosize() {
    Height = FontSize + PaddingTop + PaddingBottom;
  }

  /*
   * EVENT METHODS
   */
  public void addTextChangedListener(String methodName) { 
    registerEventMethod(5, methodName);
  }

  public void removeTextChangedListener() {
    deregisterEventMethod(5);
  }

  @Override
    protected void enter(MouseEvent e) {
    VisualBackgroundColor = HoverColor;
    update();
  }

  @Override
    protected void exit(MouseEvent e) {
    VisualBackgroundColor = BackgroundColor;
    update();
  }

  protected int selectionInitial; // cursor at the time the dragging started

  @Override
    public void press(MouseEvent e) {
    int mX = e.getX();
    int mY = e.getY();

    setCursorByClick(mX, mY);

    selectionInitial = cursorPosition;
    selectionStart = cursorPosition;
    selectionEnd = cursorPosition;

    this.Focus();
    cursorTime = papplet.millis();
    displayCurs = true;
    update();
  }

  protected void setCursorByClick(int mX, int mY) {
    int clickedPos = mX - Bounds.X0 + ScrollPosition - PaddingLeft; // relative to textbox origin and considering
    // fullScrollWidth
    float wide = 0;
    for (int i = 0; i < Text.length(); i++) {
      float letterWidth = pg.textWidth(Text.substring(i, i + 1));
      wide += letterWidth;
      if (wide - letterWidth / 2 > clickedPos) { // set decision point to the center of the letter
        moveCursorTo(i);
        break;
      }
    }
    if (wide < clickedPos) { // in case clicked beyond last letter - set cursor to end
      moveCursorTo(Text.length());
    }
  }

  @Override
    protected void drag(MouseEvent e) {
    setCursorByClick(e.getX(), e.getY());

    if (cursorPosition > selectionInitial) {
      selectionStart = selectionInitial;
      selectionEnd = cursorPosition;
    } else {
      selectionStart = cursorPosition;
      selectionEnd = selectionInitial;
    }

    // scroll a bit when at left or right edge
    if (e.getX() - Bounds.X0 < 10) {
      setScrollPositionX(ScrollPosition - 10);
    } else if (Bounds.X - e.getX() < 10) {
      setScrollPositionX(ScrollPosition + 10);
    }
  }

  boolean ctrl, shft, alt = false;

  @Override
    public void onKeyRelease(KeyEvent e) {
    switch (e.getKeyCode()) {
    case PApplet.CONTROL:
      ctrl = false;
      break;
    case PApplet.SHIFT:
      shft = false;
      break;
    case PApplet.ALT:
      alt = false;
      break;
    }
    this.Key = Key;
  }

  int Key;
  public int getKey() {
    return Key;
  }

  @Override
    public void onKeyPress(KeyEvent e) {
    if (Enabled) {
      char Key = e.getKey();
      int Code = e.getKeyCode();

      if (ctrl && !shft && !alt) {
        switch ((char) Code) {
        case 'C': 
          {
            if (selectionStart < selectionEnd) {
              StringSelection selection = new StringSelection(Text.substring(selectionStart, selectionEnd));
              Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
              clipboard.setContents(selection, selection);
            }
            break;
          }
        case 'V': 
          {
            Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
            Transferable contents = clipboard.getContents(null);
            if (contents != null && contents.isDataFlavorSupported(DataFlavor.stringFlavor)) {
              try {
                String content = (String) contents.getTransferData(DataFlavor.stringFlavor);
                if (selectionStart < selectionEnd) { // if there is a selection then replace
                  int selStart = selectionStart; // store this because delete will call cursorChanged
                  // which resets selectionStart
                  deleteRange(selectionStart, selectionEnd);
                  moveCursorTo(selStart);
                }
                this.append(content);
              } 
              catch (UnsupportedFlavorException ex) {
                System.out.println(ex);
                ex.printStackTrace();
              } 
              catch (IOException ex) {
                ex.printStackTrace();
              }
            }
          }
          break;
        case 'A':
          selectionStart = 0;
          selectionEnd = Text.length();
          update();
          break;
        case 'X':
          if (selectionStart < selectionEnd) {
            StringSelection selection = new StringSelection(Text.substring(selectionStart, selectionEnd));
            Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
            clipboard.setContents(selection, selection);
          }
          deleteRange(selectionStart, selectionEnd);
          selectionStart = 0;
          selectionEnd = 0;
          break;
        }
      }

      if (Key == PApplet.CODED) {
        switch (Code) {
        case PApplet.LEFT:

          if (shft) {
            int selEnd = selectionEnd;
            moveCursorBy(-1);
            selectionEnd = selEnd;
          } else {
            moveCursorBy(-1);
          }

          break;
        case PApplet.RIGHT:
          if (shft) {
            int selStart = selectionStart;
            moveCursorBy(1);
            selectionStart = selStart;
          } else {
            moveCursorBy(1);
          }
          break;
        case 35: // END
          moveCursorTo(Text.length());
          break;
        case 36: // pos1
          moveCursorTo(0);
          break;
        case PApplet.CONTROL:
          ctrl = true;
          break;
        case PApplet.SHIFT:
          shft = true;
          break;
        case PApplet.ALT:
          alt = true;
          break;
        }
      } else if (!ctrl) { // not coded
        switch (Key) {
        case PApplet.BACKSPACE:
          if (selectionStart < selectionEnd) {
            int selStart = selectionStart; // store this because delete will call cursorChanged which resets
            // selectionStart
            deleteRange(selectionStart, selectionEnd);
            moveCursorTo(selStart);
          } else {
            backspace();
          }
          break;
        case PApplet.DELETE:
          if (selectionStart < selectionEnd) {
            int selStart = selectionStart; // store this because delete will call cursorChanged which resets
            // selectionStart
            deleteRange(selectionStart, selectionEnd);
            moveCursorTo(selStart);
          } else {
            deleteKey();
          }
          break;
        case PApplet.ENTER:
          this.Blur();
          update();
          break;
        default:
          if (selectionStart < selectionEnd) { // if there is a selection then replace
            int selStart = selectionStart; // store this because delete will call cursorChanged which resets
            // selectionStart
            deleteRange(selectionStart, selectionEnd);
            moveCursorTo(selStart);
          }
          append(Key);
        }
      }
    }
    this.Key = Key;
  }
}
