package pGUI.core;

import processing.core.*;
import processing.event.*;

import pGUI.classes.*;

public class VScrollContainer extends VFlowContainer {

	// Width the scrollContainer would have considering all its content
	// (fullScrollHeight can actually be smaller that Height!)
	protected int fullScrollHeight;

	// Scroll position (only vertical), starts at 0
	protected int ScrollPosition = 0;

	// speed at which container will be scrolled, can be set externally
	protected int ScrollSpeed = 10;

	public VScrollContainer(PApplet pa) {
		super(pa);
		addMouseListener();
	}

	@Override
	protected void render() {
		// if AutoSize is on, first get minimal dimensions
		if (AutoSize) {
			setAutoSize();
		}

		pg = papplet.createGraphics(Width, Height);
		pg.beginDraw();
		drawDefaultBackground();

		int usedSpace = 0;

		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);

			if (c.Visible) {
				try {
					Container cx = (Container) c;
					c.Bounds.X0 = this.Bounds.X0 + c.MarginLeft + PaddingLeft;
					c.Bounds.Y0 = this.Bounds.Y0 + usedSpace + c.MarginTop + PaddingTop - ScrollPosition;

					c.Bounds.X = Math.min(c.Bounds.X0 + c.Width, this.Bounds.X);
					c.Bounds.Y = Math.min(c.Bounds.Y0 + c.Height, this.Bounds.Y);

					// constrain after computing X,Y so no data will be lost by constraining
					c.Bounds.X0 = Math.max(c.Bounds.X0, this.Bounds.X0);
					c.Bounds.Y0 = Math.max(c.Bounds.Y0, this.Bounds.Y0);
					if (c.changedVisuals) {
						c.changedVisuals = false;
						c.render();
					}
				} catch (ClassCastException cce) {
					// cce.printStackTrace();
					if (c.changedVisuals) {
						c.changedVisuals = false;
						c.render();
					}
				}
				c.Bounds.X0 = this.Bounds.X0 + c.MarginLeft + PaddingLeft;
				c.Bounds.Y0 = this.Bounds.Y0 + usedSpace + c.MarginTop + PaddingTop - ScrollPosition;

				c.Bounds.X = Math.min(c.Bounds.X0 + c.Width, this.Bounds.X);
				c.Bounds.Y = Math.min(c.Bounds.Y0 + c.Height, this.Bounds.Y);

				// constrain after computing X,Y so no data will be lost by constraining
				c.Bounds.X0 = Math.max(c.Bounds.X0, this.Bounds.X0);
				c.Bounds.Y0 = Math.max(c.Bounds.Y0, this.Bounds.Y0);

				// subtract scrollPositionX from x-drawing-value
				pg.image(c.getGraphics(), c.MarginLeft + PaddingLeft,
						usedSpace + c.MarginTop + PaddingTop - ScrollPosition);
				usedSpace += (c.Height + c.MarginTop + c.MarginBottom);
			}
		}

		//// get total virtual height of all elements lined up , elements should be
		//// rendered once first, so they can adjust their size
		fullScrollHeight = PaddingTop;
		for (int i = 0; i < Content.size(); i++) {
			Control c = Content.get(i);

			if (c.Visible) {
				fullScrollHeight += c.MarginTop + c.Height + c.MarginBottom;
			}
		}

		// Draw view-only scrollbar:
		float scrollbarHeight = (float) Height / fullScrollHeight * Height;

		if (scrollbarHeight < Height) { // don't display scroll-bar when there's nothign to scroll

			float scrollBarY = PApplet.constrain(
					ScrollPosition * (Height - scrollbarHeight) / (fullScrollHeight - Height), 1,
					Height - scrollbarHeight - 2);
			pg.fill(150);
			pg.noStroke();
			pg.rect(Width - 4, scrollBarY, 3, scrollbarHeight, 15);
		}

		pg.endDraw();
	}

	public void setScrollPositionY(float i) {
		ScrollPosition = PApplet.constrain((int) i, 0, PApplet.max(0, fullScrollHeight - Height));
		update();
	}

	/*
	 * SETTER
	 */

	public void setScrollSpeed(int s) {
		ScrollSpeed = s;
	}

	public int getScrollSpeed() {
		return ScrollSpeed;
	}

	/*
	 * MOUSE EVENTS
	 */
	@Override
	protected void mouseWheel(MouseEvent e) {
		setScrollPositionY(ScrollPosition + e.getCount() * ScrollSpeed);
	}
}