/*
 * Main class for the eye tracking process.
 *
 * Most stuff is managed in here, like settings up the FeatureDetection (uses (not anymore) OpenCV to detect rough position of eyes and the nose) 
 *   reading the image data, leading the user through the calibration process and finally evaluating the calculated data and finding out the direction the user looks. 
 * 
 */





import java.awt.Rectangle;



class EyeTracker {
  public int tL = 8; //tolerance for left
  public int tR = 8; //.. right
  public int tU = 7; //.. up
  public int tD = 6; //.. down

  boolean calibrating;
  int calibratingStage;

  float irisX_0 = 0; //average x-Coord of eye; /relative to image origin
  float irisY_0 = 0; //"" y; relative to image origin
  int irisX; //iris position, relative to image origin, result of each hough-transform
  int irisY;
  int irisDiameter; 

  public Rectangle feature;

  private int imageWidth;
  private int imageHeight;
  private int imageX; //coordinates for top-left corner of live camera image in the screen
  private int imageY; 

  private int contourWhiteReflectionThreshold = 120;
  private int originalBlackPupilThreshold = 70;

  //public PGraphics HoughOutputPlane;
  private PGraphics output;
  public static final int a = 0;

  color bgcolor;


  private FeatureDetection fd;
  private GData gg;
  CamContainer cc;

  PApplet parent;

  int calibrationVersion = 2;



  EyeTracker(PApplet parent, CamContainer cc) {
    this.parent = parent;
    this.cc = cc;

    this.imageWidth = cc.cam.width;
    this.imageHeight = cc.cam.height;
    this.imageX = cc.imgCenterX;
    this.imageY = cc.imgCenterY;

    //this.gg = new GData(parent, cc.cam.width, cc.cam.height);
    gg = new GData();
    this.fd = new FeatureDetection(parent, EYE, this.cc.w, this.cc.h);
    this.fd.setPosition(cc.imgCenterX, cc.imgCenterY);
    //this.HoughOutputPlane = createGraphics(this.imageWidth, this.imageHeight);
    this.output = createGraphics(imageWidth, imageHeight);

    this.calibrating = true;
    this.calibratingStage = 1;


    referenceDot = new Circle();
    currentIris = new Circle();
    result = new Circle();
  }








  void calibrate() { //main calibration method
    if (calibrationVersion == 1) {
      switch(this.calibratingStage) {
      case 0: 
        calibrateCam(); //first calibrate light of cam
        break;
      case 1: 
        calibrateFD(); //then look for eyes/faces
        break;
      case 2: 
        calibrateHough(); //ajust eye-diameter size
        break;
      case 3: 
        calibrateEyePosition();
        break;
      default: 
        endCalibration();
        break;
      }
    } else if (calibrationVersion == 2) {
      switch(calibration2Stage) {
      case 0: //stage 0
        findReferenceDot();
        break;
      case 1: //stage 1
        findEye(0); //0 = left eye
        break;
      case 2: //
        testIrisDiameter();
        break;
      case 3: //stage 3
        calibrateIrisLocation();
        break;
      case 4:
        endCalibrationVersion2();
      }
    }
  }

  int calibration2Stage = 0;
  void resetCalibration2() {
    calibration2Stage = 0;
    irisX_0 = 0;
    irisY_0 = 0;
    calibrating = true;
    calIrisLoc_Step = 0;
    findEye_Step = 0;
    currentIris = new Circle();
    referenceDot = new Circle();
  }

  int goUpToCalib;
  int leaveOutFirstImages = 0;



  Circle referenceDot;  //current location and diameter of the reference dot (relative to image origin)
  Circle currentIris;   //current location and diameter of the iris (relative to image origin)

  int relationX;        //const. calibrated difference between reference dot and iris (x-Direction)
  int relationY;        //const. calibrated difference between reference dot and iris (y-Direction)

  Circle result;        //current difference between reference dot and iris location, considering relationX and Y.
  //when x and y of this variable is 0, then the user is looking at the center





  /* 
   * STAGE 0
   * 
   * First find the reference, pinned to the forehead of the user. 
   * Especially get the diameter of it.
   */
  int goToUp;
  void findReferenceDot() { 

    if (leaveOutFirstImages < 10) { //ignore first image from camera, cause the colors aren't adjusted yet, thus creating high peaks in the red spectrum. This slows down following algorithm immensly
      leaveOutFirstImages ++;
    } else {
      int averageX = 0;
      int averageY = 0;
      int averageDiameter = 0;
      int repeats = 4;

      for (int i = 0; i < repeats; i++) {
        goDownToRed++;
        feature = new Rectangle(60, 60, imageWidth-120, imageHeight-120);
        background(calibBGColor);
        output.beginDraw();


        gg.setImg(cc.cam);
        //output.set(0, 0, this.gg.uneditedImg);
        output.set(0, 0, this.gg.img);

        //referenceDot = greenCircles(this.feature.x, this.feature.y, this.feature.width+this.feature.x, this.feature.height+this.feature.y, 20, 100, 75);

        this.gg.contour(1); //der läuft für diese Anwendung super-stabil




        //output.set(0, 0, this.gg.uneditedImg);
        qqqqqqqq++;
        referenceDot = houghRed(this.feature.x, this.feature.y, this.feature.width+this.feature.x, this.feature.height+this.feature.y, 20, 100, 2);



        averageX += referenceDot.x;
        averageY += referenceDot.y;
        averageDiameter += referenceDot.diameter;

        output.endDraw();
        image(output, imageX, imageY);
        if (debugMode) {
          int clr = gg.uneditedImg.pixels[constrain((mouseX-imageX)+(mouseY-imageY)*imageWidth, 0, gg.uneditedImg.pixels.length-1)];
          textSize(12);
          text((int)red(clr) + "," + (int)green(clr)  + "," + (int)blue(clr), 20, 20);
          text("Brightness: " + (int)brightness(clr), 20, 60);
          text("Saturation: " + (int)saturation(clr), 20, 100);
          text("hue: " + (int)hue(clr), 20, 140);
        }
      }
      averageX/= repeats;
      averageY /= repeats;
      averageDiameter /= repeats;

      referenceDot.x = averageX;
      referenceDot.y = averageY;
      referenceDot.diameter = averageDiameter;
      if (!keyPressed) {
        calibration2Stage++;
      }
    }
  }



  /*
   * STAGE 1
   * 
   * Now approximate the position of the eyes from the information about location and diameter of the reference point
   */
  int findEye_Step = 0;
  Circle eyeApprox = new Circle();

  void findEye(int whichEye) { //stage 0

    float diameterRelation = 1;

    if (findEye_Step == 0) {
      //eyeApprox.x = (int)(referenceDot.x + 2.2 * referenceDot.diameter * (whichEye == 0 ? -1 : 1));
      //eyeApprox.y = referenceDot.y + 2 * referenceDot.diameter;
      eyeApprox.x = (int)(referenceDot.x + 0 * referenceDot.diameter * (whichEye == 0 ? -1 : 1));
      eyeApprox.y = referenceDot.y - 2 * referenceDot.diameter;
      eyeApprox.diameter = (int)(diameterRelation * referenceDot.diameter);

      output.beginDraw();
      output.ellipse(eyeApprox.x, eyeApprox.y, eyeApprox.diameter, eyeApprox.diameter);
      output.endDraw();
      image(output, imageX, imageY);
    } else if (findEye_Step == 1) {

      delay(1500);
    } else {

      background(calibBGColor);

      output.beginDraw();
      gg.setImg(cc.cam);

      //this.gg.contour(1);
      int t0 = millis();
      this.gg.advancedContour(feature.x-50, feature.y-50, feature.x + feature.width+50, feature.y+feature.height+50);
      this.gg.contrast(12, 120);
      timeKeeper.add(millis()-t0, "gAlgos");
      //gg.lowPass(gg.param2);
      //this.gg.advancedContour(feature.x-50, feature.y-50, feature.x + feature.width+50, feature.y+feature.height+50);

      output.set(0, 0, this.gg.img);
      output.rect(eyeApprox.x - umgebung*eyeApprox.diameter, eyeApprox.y - umgebung*eyeApprox.diameter, 2*umgebung*eyeApprox.diameter, 2*umgebung*eyeApprox.diameter);
      output.endDraw();
      currentIris = hough((int)(eyeApprox.x - umgebung*eyeApprox.diameter), (int)(eyeApprox.y - umgebung*eyeApprox.diameter), (int)(eyeApprox.x + umgebung*eyeApprox.diameter), 
        (int)(eyeApprox.y + umgebung*eyeApprox.diameter), (int)(eyeApprox.diameter), (int)(eyeApprox.diameter), 1);
      //println(currentIris.x, currentIris.y, currentIris.diameter);
      image(output, imageX, imageY);

      if (findEye_Step == 20) {
        calibration2Stage += 2;
      }
    }
    calIrisLoc_Step = 0;
    findEye_Step++; //den Schritt "testIrisDiameter" überspringen
  }
  float umgebung = 1.5;



  /*
   * STAGE 2
   * 
   * Gget the the diameter of the iris. this will test about +-30% of the reference-dot diameter
   */
  void testIrisDiameter() {
  }

  /*
   * STAGE 3
   * 
   * Measure the distance between reference-dot and iris -> set relationX, relationY
   */

  int calIrisLoc_Step = 0;
  int calIrisLoc_Repeats = 2;   //number of values taken for calibration
  int calIrisLoc_Frequency = 7;  //in times per seconds
  int calIrisLoc_CountValues;    //just to be on the save side, in case not the exact number of repeats will be executed

  void calibrateIrisLocation() {
    if (calIrisLoc_Step == 0) {
      background(calibBGColor);
      text("Look to the center\nof the screen!", 50, 0, 40, textColor);
      colorMode(RGB, 255);
      fill(100, 100, 100, 50);
      ellipse(width/2, height/2, 100, 100);
      relationX = 0;
      relationY = 0;

      calIrisLoc_CountValues = 0;
    } else if (calIrisLoc_Step == 1) {

      delay(1000);
      frameRate(60);    //have a smooth animation here

      //
    } else if (calIrisLoc_Step < 60 * calIrisLoc_Repeats / calIrisLoc_Frequency) {
      //take a value, but not every frame

      output.beginDraw();
      output.set(0, 0, this.gg.uneditedImg);
      int t0 = millis();
      gg.setImg(cc.cam);
      gg.lowPass(gg.param2);  
      this.gg.advancedContour(feature.x-50, feature.y-50, feature.x + feature.width+50, feature.y+feature.height+50);
      timeKeeper.add(millis()-t0, "LP+ADC");

      Circle iris = new Circle();

      if (calIrisLoc_Step % (60 / (calIrisLoc_Frequency)) == 0) {

        referenceDot = houghRed((int)(referenceDot.x - umgebung * referenceDot.diameter), (int)(referenceDot.y - umgebung * referenceDot.diameter), 
          (int)(referenceDot.x + umgebung * referenceDot.diameter), (int)(referenceDot.y + umgebung * referenceDot.diameter), 
          referenceDot.diameter, referenceDot.diameter, 1);

        if (currentIris != null) {
          iris = hough(currentIris.x - 2 * currentIris.diameter, currentIris.y - 2 * currentIris.diameter, 
            currentIris.x + 2 * currentIris.diameter, currentIris.y + 2 * currentIris.diameter, 
            currentIris.diameter, currentIris.diameter, 1);

          relationX += iris.x - referenceDot.x;
          relationY += iris.y - referenceDot.y;
          //println("!!!! ", iris.x, iris.y, iris.diameter);
          calIrisLoc_CountValues++;
        }
      }

      //animation
      background(bgcolor);

      output.endDraw();

      image(output, imageX, imageY);

      colorMode(RGB, 255);
      text("Look to the center\nof the screen!", 50, 0, 40, textColor);
      fill(100, 100, 100, 50);
      float d = (60*calIrisLoc_Repeats/calIrisLoc_Frequency - calIrisLoc_Step) / (float)(60*calIrisLoc_Repeats/calIrisLoc_Frequency)*width;
      ellipse(width/2, height/2, d, d);
      fill(224, 216, 0, 180);
      ellipse(width/2, height/2, 100, 100);
    } else { //finish iris location calibration
      if (calIrisLoc_CountValues > 0) {
        relationX /= calIrisLoc_CountValues;
        relationY /= calIrisLoc_CountValues;
      } else {
      }      
      calibration2Stage++;
    }
    calIrisLoc_Step++;
  }


  void endCalibrationVersion2() {
    calibrating = false;
  }




  int i = 0;
  void track2() {

    output.beginDraw();
    output.clear();

    if (liveUneditedVideo && !liveOutput) {
      set(0, 0, cc.cam);
    }

    //copy image from webcam to grahpic-editing object 
    int t0 = millis();
    this.gg.setImg(cc.cam);
    //gg.lowPass(gg.param2);
    this.gg.advancedContour(feature.x-50, feature.y-50, feature.x + feature.width+50, feature.y+feature.height+50);
    /*this.gg.advancedContour((int)(referenceDot.x - umgebung * referenceDot.diameter), (int)(referenceDot.y - umgebung * referenceDot.diameter), 
      (int)(referenceDot.x + umgebung * referenceDot.diameter), (int)(referenceDot.y + umgebung * referenceDot.diameter));
    this.gg.advancedContour((int)(referenceDot.x - umgebung * referenceDot.diameter) + relationX, (int)(referenceDot.y - umgebung * referenceDot.diameter) + relationY, 
      (int)(referenceDot.x + umgebung * referenceDot.diameter + relationX), (int)(referenceDot.y + umgebung * referenceDot.diameter + relationY));*/
    timeKeeper.add(millis()-t0, "gAlgos");

    if (liveOutput) {
      output.set(0, 0, this.gg.img);
    }

    output.line(referenceDot.x, referenceDot.y, referenceDot.x + relationX, referenceDot.y + relationY);
    Circle redResult = houghRed((int)(referenceDot.x - umgebung * referenceDot.diameter), (int)(referenceDot.y - umgebung * referenceDot.diameter), 
      (int)(referenceDot.x + umgebung * referenceDot.diameter), (int)(referenceDot.y + umgebung * referenceDot.diameter), 
      referenceDot.diameter, referenceDot.diameter, 1);

    if (redResult.x != 0) {
      referenceDot = redResult;
    }

    /*currentIris = hough(currentIris.x - 2 * currentIris.diameter, currentIris.y - 2 * currentIris.diameter, 
     currentIris.x + 2 * currentIris.diameter, currentIris.y + 2 * currentIris.diameter, 
     currentIris.diameter, currentIris.diameter, 1);*/

    output.rect(referenceDot.x - umgebung*referenceDot.diameter, referenceDot.y - umgebung*referenceDot.diameter, umgebung*2*referenceDot.diameter, umgebung*2*referenceDot.diameter);

    output.rect(referenceDot.x - umgebung*referenceDot.diameter+relationX, referenceDot.y - umgebung*referenceDot.diameter+relationY, umgebung*2*referenceDot.diameter, umgebung*2*referenceDot.diameter);

    currentIris = hough((int)(referenceDot.x - umgebung * referenceDot.diameter) + relationX, (int)(referenceDot.y - umgebung * referenceDot.diameter) + relationY, 
      (int)(referenceDot.x + umgebung * referenceDot.diameter + relationX), (int)(referenceDot.y + umgebung * referenceDot.diameter + relationY), 
      referenceDot.diameter, referenceDot.diameter, 1);
    if (currentIris != null) {
      result.x = currentIris.x-referenceDot.x - relationX;
      result.y = currentIris.y-referenceDot.y - relationY;
    } else {
      result.x = 0;
      result.y = 0;
    }
    if (currentIris.x == -2000) {
      result.x = 0;
      result.y = 0;
    }

    //text(result.y +" " +str(currentIris.y)+ " " +referenceDot.y + " " +relationY, 30, 90);
    //text(result.x +" " +str(currentIris.x) + " " +referenceDot.x  + " " +relationX, 30, 50);
    output.endDraw();

    if (liveOutput && !liveUneditedVideo) {
      image(output, imageX, imageY);
    }
  }

  int evaluate2() {
    int dirX = 1;
    int dirY = 1;


    //first calculate the resulting direction out of the data
    { 
      colorMode(RGB, 255);
      fill(0, 150, 200);
      strokeWeight(0);
      if (this.result.x > this.tL) { //left
        dirX = 0;
      }
      if (this.result.x < - this.tR) { //right
        dirX = 2;
      }
      if (this.result.y < -this.tU) { //up
        dirY = 0;
      }
      if (this.result.y > this.tD) { //down
        dirY = 2;
      }
    }

    int dir = 3*dirY + dirX;



    //only react if same direction has been measured several times
    //that should smooth out some of the jittering from hough-transform and OpenCV

    if (etMeasure > 1) { //in neue richtung gucken
      etPreviousDirX = dirX;
      etPreviousDirY = dirY;
      //interpretEyeData(dir, false);
    } 

    //draw a rect to indicate the direction
    if (debugMode) {
      rect(etPreviousDirX / 2. * (width-100), etPreviousDirY / 2. *(height-100), 100, 100);
    }

    if (etPreviousDir == dir) {
      etMeasure++;
    } else {
      etMeasure = 0;
    }
    etPreviousDir = dir;
    return dir;
  }





  void calibrateCam() { //0
    background(bgcolor);

    output.beginDraw();
    output.set(this.imageX, this.imageY, this.cc.cam);
    output.endDraw();
    image(output, 200, 200);

    fill(max(calibBGColor-20, 0), 120);
    rect(0, 0, 500, 150);
    fill(textColor);
    textSize(40);
    text("Calibrating Camera", 20, 50);
    textSize(20);
    text("Press enter to resume", 20, 90);
  }






  void calibrateFD() { //calibratingState = 1
    background(bgcolor);

    output.beginDraw();
    output.set(0, 0, this.cc.cam);
    Rectangle[] noses = fd.detectNose(cc);
    for (int i = 0; i < noses.length; i++) {
      rect(noses[0], output, false, 180);
    }
    output.endDraw();
    image(output, imageX, imageY);

    this.fd.detect(true, this.cc, this.calibratingStage);

    fill(max(calibBGColor-20, 0), 120);
    rect(0, 0, 500, 150);
    text("Searching for Eyes...", 40, 20, 50, textColor);
    text("Scroll to change detection tolerance", 20, 20, 90, textColor);
    text("Select one Eye...", 20, 20, 120, textColor);
    rect(width-40, this.fd.getTolerance(), 40, 25);
    textSize(14);
    text(this.fd.getTolerance(), width-60, this.fd.getTolerance()+20);
  }






  void calibrateHough() { //calibratingState = 2
    background(bgcolor);
    output.beginDraw();
    gg.setImg(cc.cam);
    gAlgorithms();
    output.set(0, 0, this.gg.img);
    output.rect(this.feature.x+cc.imgCenterX, this.feature.y+cc.imgCenterY, this.feature.width, this.feature.height);
    output.endDraw();

    hough(this.feature.x, this.feature.y, this.feature.width+this.feature.x, this.feature.height+this.feature.y, (int)this.gg.param2, (int)this.gg.param2, 2);

    image(output, imageX, imageY);
    //image(HoughOutputPlane, imageX, imageY);
  }








  //find average eye position, when looked to center of the screen; needs several runs!
  //__:
  int caliTime0 = 1; //time to calibrate in seconds
  int caliTime = caliTime0*generalFrameRate; //current calibrate countdwon number

  float averageFeatureX = 0; //average center x-Coord of eye-rectangle (openCV), is ...relativ... to image origin
  float averageFeatureY = 0; //"" center y, is ...relative... to image origin
  float averageFeatureCountAddedValues = 0; //count how many values are being added to "averageFeatureX/Y" (Sometimes openCV doesn't detect any eyes).
  int c = 0;
  int rel = 1;
  int calitime2 = caliTime0 * generalFrameRate * rel;


  void calibrateEyePosition() { //calibratingState = 3
    if (caliTime0 * generalFrameRate == caliTime) {
      background(calibBGColor);
      irisDiameter = gg.param2;

      text("Look to \nthe center of the screen!", 50, 0, 40, textColor);
      textSize(20);
      fill(200, 0, 0);
      ellipse(width/2, height/2, 100, 100);
    }

    if (caliTime0 * generalFrameRate - 1 == caliTime) {
      delay(1000);
    }

    if (caliTime > 0) {
      background(calibBGColor);
      output.beginDraw();
      output.clear();

      Rectangle[] noses = fd.detectNose(cc);
      if (noses.length > 0) {
        Rectangle currentFeature = noses[0];//fitRightRectangle(noses, feature, fittingTolerance, threshold);
        if (currentFeature != null) {
          drawRectNoFill(noses[0], output);
          averageFeatureX += getCenterOfRect(noses[0]).x;
          averageFeatureY += getCenterOfRect(noses[0]).y;
          averageFeatureCountAddedValues++;
        }
      }

      this.gg.setImg(cc.cam);
      gAlgorithms();
      set(cc.imgCenterX, cc.imgCenterY, gg.img);
      hough(feature.x, feature.y, feature.width + feature.x, feature.height + feature.y, (int)gg.param2, (int)gg.param2, 2);

      output.endDraw();
      image(output, imageX, imageY);
      irisX_0 += irisX;
      irisY_0 += irisY;

      colorMode(RGB, 255);
      text("Look to the center of the screen!", 50, 0, 40, textColor);
      fill(100, 100, 100, 50);
      float d = (caliTime0 * generalFrameRate - caliTime) / (float)(caliTime0 * generalFrameRate)*width;
      ellipse(width/2, height/2, d, d);
      //fill(200, 0, 0);
      fill(224, 216, 0, 180);
      ellipse(width/2, height/2, 100, 100);
      caliTime--;
    } else { 
      //finish 
      irisX_0 /= caliTime0 * generalFrameRate;
      irisY_0 /= caliTime0 * generalFrameRate;
      averageFeatureX /= averageFeatureCountAddedValues;
      averageFeatureY /= averageFeatureCountAddedValues;
      relX = irisX_0 - averageFeatureX;
      relY = irisY_0 - averageFeatureY;

      text(averageFeatureX, 60, 60);
      calibratingStage++;
    }
  }






  float relX, relY; //distance between iris center and nose corner
  Rectangle nose;

  Rectangle fitRightRectangle(Rectangle[] array, Rectangle target, float fittingTolerance, int threshold) {
    int compareValue[] = new int[array.length+2]; //array to store the "quality" of a feature; determined by how close it is to the originally selected feature
    for (int i = 0; i < array.length; i++) { 
      if (abs(target.x - array[i].x) < fittingTolerance)
        compareValue[i]++;
      if (abs(target.y - array[i].y) < fittingTolerance)
        compareValue[i]++;
      if (abs(target.width - array[i].width) < fittingTolerance)
        compareValue[i]++;
      if (abs(target.height - array[i].height) < fittingTolerance)
        compareValue[i]++;
    }
    //get highest value
    int maxCompareValue = 0;
    int maxCompareValueIndex = -1;
    for (int i = 0; i < compareValue.length; i++) {
      if (compareValue[i] > maxCompareValue) {
        maxCompareValue = compareValue[i];
        maxCompareValueIndex = i;
      }
    }
    if (maxCompareValue > threshold)
      return array[maxCompareValueIndex];
    else return null;
  }

  void endCalibration() {
    calibrating = false;
  }

  void resetCalib() {
    irisX_0 = 0;
    irisY_0 = 0;
    averageFeatureX = 0; //average x-Coord of eye-rectangle (openCV), is ...relativ... to image origin
    averageFeatureY = 0; //"" y, is ...relative... to image origin
    averageFeatureCountAddedValues = 0;
    calibratingStage = 1;
    calibrating = true;
    caliTime = caliTime0*generalFrameRate;
    calitime2 = caliTime0*generalFrameRate*rel;
  }


  void resetRelationCalib() {
    irisX_0 = 0;
    irisY_0 = 0;
    averageFeatureX = 0; //average x-Coord of eye-rectangle (openCV), is ...relativ... to image origin
    averageFeatureY = 0; //"" y, is ...relative... to image origin
    averageFeatureCountAddedValues = 0;
    calibratingStage = 3;
    calibrating = true;
    caliTime = caliTime0*generalFrameRate;
    calitime2 = caliTime0*generalFrameRate*rel;
  }







  void gAlgorithms() {

    //myG.simpleBlur();
    //myG.lowPass(myG.param2);
    int t0 = millis();
    //this.gg.blur();
    this.gg.advancedContour(feature.x-50, feature.y-50, feature.x + feature.width+50, feature.y+feature.height+50);
    //this.gg.contour(1);
    //this.gg.advancedContour(0, 0, gg.w, gg.h);
    //myG.lowPass(myG.param2);
    //myG.threshold(myG.param1);
    //gg.optimize();
    //gg.simpleBlur();
    //println("gg-algos-time:", millis()-t0);
  }








  public void track() {
    output.beginDraw();
    output.clear();

    if (liveUneditedVideo && !liveOutput) {
      set(0, 0, cc.cam);
    }

    //copy image from webcam to grahpic-editing object 
    this.gg.setImg(cc.cam);


    //apply graphic algorithms
    gAlgorithms();


    if (liveOutput) {
      output.set(0, 0, this.gg.img);
    }

    //search for circles
    hough(feature.x, feature.y, feature.width + feature.x, feature.height + feature.y, irisDiameter, irisDiameter, 2);

    //evaluate result coordinates (finalX, finalY) of hough-transform 
    //evaluate();

    //get nose position
    /*Rectangle[] noses = fd.detectNose(cc); 
     if (noses.length > 0) {
     //irisX_0 = getCenterOfRect(noses[0]).x + relX;
     //irisY_0 = getCenterOfRect(noses[0]).y + relY;
     }*/
    output.endDraw();
    image(output, imageX, imageY);

    //if (detectFeaturesLive && ++observeDetectionFrequency % (generalFrameRate/faceDetectionFrequency) == 0 || true) { }
    if (liveOutput) {


      /*if (noses.length > 0)
       drawRectNoFill(noses[0], output);*/
      output.endDraw();
      image(output, imageX, imageY);
    }
    //evaluate();
  }


  int etMeasure = 1;       //count the number of follwing same directions
  int etPreviousDir;       //Direction of previous frame
  int etPreviousDirX = 1;  //dirX of prev frame
  int etPreviousDirY = 1;  //dirY of prev frame

  public int evaluate() {
    int dirX = 1;
    int dirY = 1;


    //first calculate the resulting direction out of the data
    { 
      colorMode(RGB, 255);
      fill(0, 150, 200);
      strokeWeight(0);
      if (this.irisX - (this.irisX_0) > this.tL) { //left
        dirX = 0;
      }
      if (this.irisX - (this.irisX_0) < - this.tR) { //right
        dirX = 2;
      }
      if (this.irisY - (this.irisY_0) < -this.tU) { //up
        dirY = 0;
      }
      if (this.irisY - (this.irisY_0) > this.tD) { //down
        dirY = 2;
      }
    }

    int dir = 3*dirY + dirX;



    //only react if same direction has been measured several times
    //that should smooth out some of the jittering from hough-transform and OpenCV

    if (etMeasure > 1) { //in neue richtung gucken
      etPreviousDirX = dirX;
      etPreviousDirY = dirY;
      //interpretEyeData(dir, false);
    } 

    //draw a rect to indicate the direction
    rect(etPreviousDirX / 2. * (width-100), etPreviousDirY / 2. *(height-100), 100, 100);

    if (etPreviousDir == dir) {
      etMeasure++;
    } else {
      etMeasure = 0;
    }
    etPreviousDir = dir;
    return dir;
  }




  public void setToleranceL(int newTolerance) {
    if (newTolerance >= 0) {
      this.tL = newTolerance;
    }
  }
  public void setToleranceR(int t) {
    if (t >= 0) {
      this.tR = t;
    }
  }
  public void setToleranceU(int t) {
    if (t >= 0) {
      this.tU = t;
    }
  }
  public void setToleranceD(int t) {
    if (t >= 0) {
      this.tD = t;
    }
  }









  Circle hough(int x0, int y0, int xMax, int yMax, int rMin, int rMax, int rasterResolution) { //d:circle diameter, r:raster resolution
    //intializing
    //int t0 = millis(); //for time performance measurement, get starting time
    xMax = floor((xMax-x0)/rasterResolution)*rasterResolution + x0; //set xMax to a multiple of rasterResolution
    output.beginDraw();
    output.noFill();
    output.strokeWeight(1);
    output.stroke(200);
    output.rect(x0, y0, xMax-x0, yMax-y0);
    if (irisX_0 != 0) {
      output.ellipse(irisX_0, irisY_0, irisDiameter, irisDiameter);
    }

    int numberHits = 1;                    //determines how many maxima will be searched

    IntList cValues = new IntList();       //contains the results of the hough-operation
    int cMaxs[] = new int[numberHits];     //array for the 5 highest values
    int cMaxsPos[] = new int[numberHits];  //array for the position if the same 5 highest values
    IntList posX = new IntList();          //contains x position of value
    IntList posY = new IntList();          //contains y position of value
    IntList rx = new IntList();            //contains radius of value
    int t0 = millis();
    int wid = cc.cam.width;


    int houghMinContrast = 0;


    //EVENTUELL SPÄTER NUR NACH HALBKREISEN SUCHEN
    for (int r = rMin; r <= rMax; r++) {
      for (int y = y0; y < yMax; y += rasterResolution) {
        for (int x = x0; x < xMax; x += rasterResolution) {
          if (brightness(gg.img.pixels[constrain(y * wid + x, 0, gg.img.pixels.length-1)]) > contourWhiteReflectionThreshold || brightness(gg.uneditedImg.pixels[constrain(y * wid + x, 0, gg.img.pixels.length-1)]) < originalBlackPupilThreshold) { //überprüft punkt nur wenn Mittelpunkt des Kreises weiß ist. Funktioniert nur unter der Prämisse, dass die Pupille teilweise weiß ist (Reflexion des Bildschirms)
            int value = 0;
            for (int i = 0; i < 360; i += 5) {
              int xx = (int)(r/2*cos(radians(i)))+x;
              int yy = (int)(r/2*sin(radians(i)))+y;
              if (yy*wid+xx >= 0 && yy*wid+xx+2 < gg.img.pixels.length) {
                /*if(brightness(gg.img.pixels[yy*wid+xx])>houghMinContrast){
                 value += 1;
                 }*/
                value += brightness(gg.img.pixels[yy*wid+xx]);
                value += brightness(gg.img.pixels[yy*wid+xx+2]); //make radius quasi-wider (here only in x-direction), so thicker circles can be better recognized
              }
            }
            cValues.append(value);
            posX.append(x);
            posY.append(y);
            rx.append(r);
          }
        }
      }
    }
    //println("HOUGH: " + (millis()-t0));

    //get max values
    for (int k = 0; k < min(cMaxs.length, cValues.size()); k++) {
      cMaxs[k] = 0;
      for (int j = 0; j < cValues.size(); j++) {
        if (cValues.get(j) > cMaxs[k]) {
          cMaxs[k] = cValues.get(j);
          cMaxsPos[k] = j;
        }
      }
      cValues.set(cMaxsPos[k], 0);
    }
    text(cMaxs[0], width/2, 3*height/4);

    output.colorMode(HSB, 255);
    output.stroke(200, 200, 200);
    int clrCircle = 255; //circle-color, so highest accuracies are marked more red, lower rather blue
    irisDiameter = rMin;

    Circle result = new Circle();

    for (int k = 0; k < min(numberHits, cValues.size()); k++) {
      output.stroke(clrCircle, 200, 200);
      clrCircle -=50;
      result.diameter = rx.get(cMaxsPos[k]);
      irisDiameter = result.diameter;
      output.ellipse(posX.get(cMaxsPos[k]), posY.get(cMaxsPos[k]), irisDiameter, irisDiameter);
    }
    if (posX.size() > 0) {
      //println("Hough - "+cValues.size()+" values, max: "+cMaxs[0] +"("+posX.get(cMaxsPos[0]), posY.get(cMaxsPos[0])+"), r: " +diameter+", t=" + (millis()-t0));
      result.x = posX.get(cMaxsPos[0]);
      irisX = result.x;
      result.y = posY.get(cMaxsPos[0]);
      irisY = result.y;
    } else { 
      //println("no values");
      result.y = -2000;
      result.x = -2000;
    }
    output.endDraw();
    timeKeeper.add(millis()-t0, "Hough");
    return result;
  }








  int goDownToRed;


  Circle houghRed(int x0, int y0, int xMax, int yMax, int rMin, int rMax, int rasterResolution) { //d:circle diameter, r:raster resolution
    goUpToCalib++;

    xMax = floor((xMax-x0)/rasterResolution)*rasterResolution + x0; //set xMax to a multiple of rasterResolution
    output.beginDraw();
    output.noFill();
    output.strokeWeight(1);
    output.stroke(200);
    output.rect(x0, y0, xMax-x0, yMax-y0);

    int numberHits = 1;                    //determines how many maxima will be searched

    IntList cValues = new IntList();       //contains the results of the hough-operation
    int cMaxs[] = new int[numberHits];     //array for the 5 highest values
    int cMaxsPos[] = new int[numberHits];  //array for the position if the same 5 highest values
    IntList posX = new IntList();          //contains x position of value
    IntList posY = new IntList();          //contains y position of value
    IntList rx = new IntList();            //contains radius of value
    int t0 = millis();
    int wid = cc.cam.width;

    int redMinContrast = 150;
    output.stroke(color(0));

    //EVENTUELL SPÄTER NUR NACH HALBKREISEN SUCHEN
    for (int y = y0; y < yMax; y += rasterResolution) {
      for (int x = x0; x < xMax; x += rasterResolution) {

        int clr  = gg.uneditedImg.pixels[constrain(y * wid + x, 0, gg.img.pixels.length-1)];

        if (hue(clr) > hue - 25 && hue(clr) < hue + 25 && saturation(clr) > sat && brightness(clr) > bri) {

          //if (green(pixelColor) > (blue(pixelColor) + red(pixelColor))/1 && brightness(pixelColor) > 100 && saturation(pixelColor)>50) {
          output.point(x, y);
          for (int r = rMin; r <= rMax; r++) {
            output.point(x, y);
            int value = 0;
            for (int i = 0; i < 360; i += 5) {
              int xx = (int)(r/2*cos(radians(i)))+x;
              int yy = (int)(r/2*sin(radians(i)))+y;
              if (yy*wid+xx >= 0 && yy*wid+xx+2 < gg.img.pixels.length) {
                /*if(brightness(gg.img.pixels[yy*wid+xx])>redMinContrast){
                 value += 1;
                 }*/
                value += brightness(gg.img.pixels[yy*wid+xx]);
                value += brightness(gg.img.pixels[yy*wid+xx+2]); //make radius quasi-wider (here only in x-direction), so thicker circles can be better recognized
              }
            }
            cValues.append(value);
            posX.append(x);
            posY.append(y);
            rx.append(r);
          }
        }
      }
    }

    //get max values
    for (int k = 0; k < min(cMaxs.length, cValues.size()); k++) {
      cMaxs[k] = 0;
      for (int j = 0; j < cValues.size(); j++) {
        if (cValues.get(j) > cMaxs[k]) {
          cMaxs[k] = cValues.get(j);
          cMaxsPos[k] = j;
        }
      }
      cValues.set(cMaxsPos[k], 0);
    }

    output.colorMode(HSB, 255);
    output.stroke(200, 200, 200);
    int clrCircle = 255; //circle-color, so highest accuracies are marked more red, lower rather blue
    //irisDiameter = rx.get(cMaxsPos[0]);


    Circle result = new Circle();
    for (int k = 0; k < min(numberHits, cValues.size()); k++) {
      output.stroke(clrCircle, 200, 200);
      clrCircle -=50;
      result.diameter = rx.get(cMaxsPos[k]);
      irisDiameter = result.diameter;
      output.ellipse(posX.get(cMaxsPos[k]), posY.get(cMaxsPos[k]), irisDiameter, irisDiameter);
      //
    }
    if (posX.size() > 0) {
      //println("Hough - "+cValues.size()+" values, max: "+cMaxs[0] +"("+posX.get(cMaxsPos[0]), posY.get(cMaxsPos[0])+"), r: " +diameter+", t=" + (millis()-t0));

      result.x = posX.get(cMaxsPos[0]);
      irisX = result.x;
      result.y = posY.get(cMaxsPos[0]);
      irisY = result.y;
    } else { 
      //println("no values");
    }
    output.endDraw();
    timeKeeper.add(millis()-t0, "HoughRed");

    return result;
  }

  int qqqqqqqq;
  Circle greenCircles(int x0, int y0, int xMax, int yMax, int rMin, int rMax, int hue) {
    goToUp++;
    int t0 = millis();
    x0 = constrain(x0, rMax, gg.img.width-rMax);
    y0 = constrain(y0, rMax, gg.img.height-rMax);
    xMax = constrain(xMax, rMax, gg.img.width-rMax);
    yMax = constrain(yMax, rMax, gg.img.height-rMax);

    int aimedHue = hue;
    output.beginDraw();
    output.strokeWeight(1);
    output.stroke(color(0));

    int rows[] = new int[(yMax-y0)];
    int columns[] = new int[(xMax-x0)];
    int area[][] = new int[xMax-x0][yMax-y0];
    //search for green pixels
    for (int y = y0; y < yMax; y++) {

      for (int x = x0; x < xMax; x++) {
        int clr = gg.uneditedImg.pixels[y * gg.img.width + x];
        if (hue(clr) > aimedHue - 20 && hue(clr) < aimedHue + 20 && saturation(clr) > 50 && brightness(clr) > 50) {
          output.point(x, y);

          gg.img.pixels[x + y * gg.img.width] = color(0);

          area[x-x0][y-y0]++;
          rows[y-y0]++;
          columns[x-x0]++;
        }
      }
    }
    gg.img.updatePixels();


    //RENDER IN BLOCKS
    /*int blockNumberX = 50;
     int blockNumberY = 40;
     int blockLength = (xMax-x0)/blockNumberX;
     int blockHeight = (yMax-y0)/blockNumberY;
     int area2[][] = new int[blockNumberX+2][blockNumberX+2];
     
     for (int x=0; x < xMax-x0; x++) { 
     for (int y = 0; y < yMax - y0; y++) {
     if (x/blockLength > 49) break;
     area2[x/blockLength][y/blockHeight]+=area[x][y];
     }
     }
     for (int x=0; x < blockNumberX; x++) {
     for (int y = 0; y <blockNumberY; y++) {
     output.fill(area2[x][y]);
     //output.rect(x0+x*blockLength, y0+y*blockHeight, blockLength, blockHeight);
     }
     }*/

    int bestRow = 0;
    int maxVal = 0;
    for (int i = 0; i < yMax-y0; i++) {
      if (rows[i]>maxVal) {
        maxVal = rows[i];
        bestRow = i;
      }
    }
    bestRow += y0;
    output.line(x0, bestRow, xMax, bestRow);   

    int bestCol = 0;
    int maxValCol = 0;
    for (int i = 0; i < xMax-x0; i++) {
      if (columns[i]>maxValCol) {
        maxValCol = columns[i];
        bestCol = i;
      }
    }
    bestCol += x0;
    output.line(bestCol, y0, bestCol, yMax);  


    /*
     * Now apply hough to area around bestCol and bestRow
     */
    this.gg.contour(1);
    //output.set(0, 0, this.gg.img);
    Circle result = new Circle();
    int wid = cc.cam.width;

    //result = hough(bestCol-20, bestRow-20, bestCol+20, bestRow+20, 20, 100, 1);

    /*
    for (int r = rMin; r <= rMax; r++) {
     for (int y = bestRow - 20; y < bestRow + 20; y++) {
     for (int x = bestCol - 20; x < bestCol + 20; x++) {
     if (area[x][y] == 1) {
     
     int value = 0;
     
     for (int i = 0; i < 360; i += 5) {
     
     int xx = (int)(r/2*cos(radians(i)))+x;
     int yy = (int)(r/2*sin(radians(i)))+y;
     
     if (yy*wid+xx >= 0 && yy*wid+xx+2 < gg.img.pixels.length) {
     ///*if(brightness(gg.img.pixels[yy*wid+xx])>houghMinContrast){
     // value += 1;
     //}
     value += brightness(gg.img.pixels[yy*wid+xx]);
     //value += brightness(gg.img.pixels[yy*wid+xx+2]); //make radius quasi-wider (here only in x-direction), so thicker circles can be better recognized
     }
     }
     }
     }
     }
     }
     
     
     
     int numberHits = 1;                    //determines how many maxima will be searched
     IntList cValues = new IntList();       //contains the results of the hough-operation
     int cMaxs[] = new int[numberHits];     //array for the 5 highest values
     int cMaxsPos[] = new int[numberHits];  //array for the position if the same 5 highest values
     IntList posX = new IntList();          //contains x position of value
     IntList posY = new IntList();          //contains y position of value
     IntList rx = new IntList();            //contains radius of value
     int wid = cc.cam.width;
     
     
     int houghMinContrast = 0;
     x0 = bestCol - 20;
     y0 = bestRow - 20;
     xMax = bestCol + 20;
     yMax = bestRow + 20;
     rMin = 20;
     rMax = 100;
     output.noFill();
     output.strokeWeight(1);
     output.stroke(200);
     output.rect(x0, y0, xMax-x0, yMax-y0);
     
     //EVENTUELL SPÄTER NUR NACH HALBKREISEN SUCHEN
     for (int r = rMin; r <= rMax; r++) {
     for (int y = y0; y < yMax; y += 1) {
     for (int x = x0; x < xMax; x += 1) {
     //if (   brightness(gg.uneditedImg.pixels[constrain(y * wid + x, 0, gg.img.pixels.length-1)]) < originalBlackPupilThreshold) { //überprüft punkt nur wenn Mittelpunkt des Kreises weiß ist. Funktioniert nur unter der Prämisse, dass die Pupille teilweise weiß ist (Reflexion des Bildschirms)
     int value = 0;
     for (int i = 0; i < 360; i += 5) {
     int xx = (int)(r/2*cos(radians(i)))+x;
     int yy = (int)(r/2*sin(radians(i)))+y;
     if (yy*wid+xx >= 0 && yy*wid+xx+2 < gg.img.pixels.length) {
     //if(brightness(gg.img.pixels[yy*wid+xx])>houghMinContrast){
     // value += 1;
     // }
     value += brightness(gg.img.pixels[yy*wid+xx]);
     value += brightness(gg.img.pixels[yy*wid+xx+2]); //make radius quasi-wider (here only in x-direction), so thicker circles can be better recognized
     }
     }
     cValues.append(value);
     posX.append(x);
     posY.append(y);
     rx.append(r);
     // }
     }
     }
     }
     
     //get max values
     for (int k = 0; k < min(cMaxs.length, cValues.size()); k++) {
     cMaxs[k] = 0;
     for (int j = 0; j < cValues.size(); j++) {
     if (cValues.get(j) > cMaxs[k]) {
     cMaxs[k] = cValues.get(j);
     cMaxsPos[k] = j;
     }
     }
     cValues.set(cMaxsPos[k], 0);
     }
     text(cMaxs[0], width/2, 3*height/4);
     
     output.colorMode(HSB, 255);
     output.stroke(200, 200, 200);
     int clrCircle = 255; //circle-color, so highest accuracies are marked more red, lower rather blue
     irisDiameter = rMin;
     
     
     for (int k = 0; k < min(numberHits, cValues.size()); k++) {
     output.stroke(clrCircle, 200, 200);
     clrCircle -=50;
     result.diameter = rx.get(cMaxsPos[k]);
     irisDiameter = result.diameter;
     output.ellipse(posX.get(cMaxsPos[k]), posY.get(cMaxsPos[k]), irisDiameter, irisDiameter);
     }
     if (posX.size() > 0) {
     //println("Hough - "+cValues.size()+" values, max: "+cMaxs[0] +"("+posX.get(cMaxsPos[0]), posY.get(cMaxsPos[0])+"), r: " +diameter+", t=" + (millis()-t0));
     result.x = posX.get(cMaxsPos[0]);
     irisX = result.x;
     result.y = posY.get(cMaxsPos[0]);
     irisY = result.y;
     } else { 
     //println("no values");
     result.y = -2000;
     result.x = -2000;
     }
     */

    if (result.x == -2000) {
      output.rect(0, 0, 400, 400);
      println("fails");
    }



    output.endDraw();   
    println(millis()-t0);
    timeKeeper.add(millis()-t0, "green");
    return result;
  }



  class Circle {
    int x;
    int y;
    int diameter;
  }
}


































class CalibrationTimer extends Thread {
  private boolean active;
  private EyeTracker parent;
  private int numberMeasurements;


  CalibrationTimer(EyeTracker parent, int numberMeasurements) {
    this.parent = parent;
    active = false;
    this.numberMeasurements = numberMeasurements;
  }


  public void start() {
    if (!active) {
      active = true;
      super.start();
    }
  }


  void run() {
    while (true) {
      if (active) {
        execute();
      } else {
        break;
      }
    }
  }


  private void execute() {
    float avgNoseX = 0;
    float avgNoseY = 0;
    float avgIrisX = 0;
    float avgIrisY = 0;
    EyeTracker e = new EyeTracker(this.parent.parent, camera);

    for (int i = 0; i < numberMeasurements; i++) {
      Rectangle[] noses = e.fd.detectNose(camera);

      if (noses.length > 0) {
        Rectangle currentFeature = noses[0];//fitRightRectangle(noses, feature, fittingTolerance, threshold);
        if (currentFeature != null) {
          //drawRectNoFill(noses[0], output);
          avgNoseX += getCenterOfRect(noses[0]).x;
          avgNoseY += getCenterOfRect(noses[0]).y;
        }
      }

      //parent.gg.setImg(cc.cam);
      //parent.gAlgorithms();
      //set(cc.imgCenterX, cc.imgCenterY, gg.img);
      //parent.hough(feature.x, feature.y, feature.width + feature.x, feature.height + feature.y, (int)parent.gg.param2, (int)parent.gg.param2, 2);

      //output.endDraw();
      //image(output, imageX, imageY);
      avgIrisX += e.irisX;
      avgIrisY += e.irisY;
    }
    avgIrisX /= numberMeasurements;
    avgIrisY /= numberMeasurements;
    avgNoseX /= numberMeasurements;
    avgNoseY /= numberMeasurements;
    parent.relX = avgIrisX - avgNoseX;
    parent.relY = avgIrisY - avgNoseY;

    parent.calibratingStage++;
    quit();
  }


  public void quit() {
    active = false;
    interrupt();
  }
}
