/*
 * The Cam-Container class provides an easy-to-use interface to initialize a Capture object.
 *
 * The user is required to set the number of the camera to use it
 * Note that the application will be closed if the specified camera number is out of bounds or not set at all!
 * It is necessary to pass the PApplet (this) to the initCam method. 
 */
 
 
static StringList availableCameras = new StringList(); //array contains list of available cameras

boolean cameraListReady = false;                       //true when thread for getting camera list has finished

void captureList() {                                   //get a list of available cameras
  cameraListReady = false; 
  String[] arr = Capture.list();
  availableCameras.clear();
  for (int i = 0; i < arr.length; i++) {
    availableCameras.append(arr[i]);
  }
  cameraListReady = true;
}





import processing.video.*;

public class CamContainer {
  public Capture cam; 
  public int w;               //width of cam image
  public int h;               //height of cam image
  public int imgCenterX;      //x-coordinate for position where image has to be displayed on the screen, so the image will stay centered; 
  public int imgCenterY;      //y-coordinate -""-
  //both absolute to display!
  
  static final int a = 0;
  private int cameraNumber;   //number of camera to use (position in array above)
  private boolean camInitialized = false; 



  CamContainer() {
    this.w = 0;
    this.h = 0;
    this.cameraNumber = -1;
    this.imgCenterX = 0;
    this.imgCenterY = 0;
    if (!fastCam) {
      thread("captureList");
    }
  }



  boolean initCam(PApplet pa) {             //initializing the cam object, and getting the camera; returns true when successful
    if (cameraListReady) {
      if (this.cameraNumber > -1) {
        if (availableCameras.size() == 0) { //display error, when no camera connected; retry twice per second
          pa.textAlign(CENTER);
          pa.textSize(20);
          pa.text("No cameras connected", width/2, height/3);
          pa.textAlign(LEFT);
          //frameRate(2);
          pa.thread("captureList");            //re-search for cameras
          return false;
        } else {
          printCameraList();
          if (this.cameraNumber >= 0 && this.cameraNumber < availableCameras.size()) {
            if (!fastCam)
              this.cam = new Capture(pa, availableCameras.get(this.cameraNumber));
            else
              this.cam = new Capture(pa, "name=Microsoft® LifeCam HD-3000,size=640x480,fps=15/2");

            this.cam.start();

            while (this.cam.height == 0 && this.cam.width == 0) {
              this.cam.read();
            }
            this.w = this.cam.width;
            this.h = this.cam.height;
            this.imgCenterX = (width - this.cam.width) / 2;
            this.imgCenterY = (height - this.cam.height) / 2;


            this.camInitialized = true;
            println("Camera is ready!");
            return true;
          } else {
            println("!! Not a valid camera number");
            exit();
            return false;
          }
        }
      } else {
        println("!! Please specify a camera number to use");
        exit();
        return false;
      }
    } else return false;
  }




  void selectCamera(int cameraNumber) { //setter for cameraNumber
    this.cameraNumber = cameraNumber;
  }

  boolean isInitialized() { //check if cam-container has been initialized
    return this.camInitialized;
  }

  void printCameraList() { //print list of available cameras
    for (int i = 0; i < availableCameras.size(); i++) {
      println(i + availableCameras.get(i));
    }
  }

  StringList getCameras() { //get list available cameras
    return availableCameras;
  }
}



void captureEvent(Capture video) {
  video.read();
}
