/* 
 * A class for displaying simple popup notifactions. 
 * It will be displayed only one notification at a time overwriting information about the last. 
 *
 * Notification can be set with setNotification(String content, int duration).
 * Duration is in milliseconds - the notifaction will appear in the center of the screen and remove itself after specified time. 
 *
 */

public class Notifier {
  color fontColor;
  int fontSize;
  color background;
  String content;
  int duration; //in milliseconds
  PGraphics notificationPlane;
  boolean display = false;

  Notifier() {
    this.fontColor = 250;
    this.fontSize = 60;
    colorMode(RGB);
    this.background = color(0, 0, 0, 180);
    notificationPlane = createGraphics(width, height);
    registerMethod("draw", this);
  }

  Notifier(int fontColor, int fontSize) {
    this.fontColor = fontColor;
    this.fontSize = fontSize;
    colorMode(RGB);
    this.background = color(0, 0, 0, 180);
    notificationPlane = createGraphics(width, height);
    registerMethod("draw", this);
  }

  public void draw() {
    if (this.display) {
      resetMatrix();
      image(this.notificationPlane, 0, height/6);
    }
  }

  void setNotification(String content, int duration) {
    this.content = content.replaceAll("\n", " ");
    this.duration = duration;

    //draw notification on PGraphics
    notificationPlane.beginDraw();
    notificationPlane.fill(background);
    notificationPlane.textSize(fontSize);
    notificationPlane.textAlign(CENTER, CENTER);

    float rectWidth = notificationPlane.textWidth(this.content) * 1.2;
    float rectHeight = fontSize * 1.8;
    float posX = width/2;
    float posY = height/2;

    notificationPlane.noStroke();
    notificationPlane.rect(posX-rectWidth/2, posY-rectHeight/2, rectWidth, rectHeight);
    notificationPlane.fill(fontColor);
    notificationPlane.text(this.content, posX, posY-5);
    notificationPlane.endDraw();
    display = true;

    NotificationTimer timer = new NotificationTimer(this);
    timer.start(duration);
  }
}









class NotificationTimer extends Thread {
  private boolean active;
  private int time;
  private Notifier parent;


  NotificationTimer(Notifier parent) {
    this.parent = parent;
    active = false;
  }


  public void start(int time) {
    if (!active) {
      active = true;
      this.time = time;
      super.start();
    }
  }


  void run() {
    while (true) {
      if (active) {
        execute();
      } else {
        break;
      }
    }
  }


  private void execute() {
    delay(time);
    quit();
    parent.display = false;
    parent.notificationPlane.clear();
  }


  public void quit() {
    active = false;
    interrupt();
  }
}
