/*
 *  Tiny task manager class, just for debugging.
 *  Times for algorithms can be added with timeKeeper.add(int time, String name); and will
 *    be displayed on a secondary window
 *
 *  Useful for keeping an eye on long algorithms and cpu-intensive methods
 */



TimeKeeper timeKeeper;

void initTimeKeeper() {
  String[] args = {"TwoFrameTest"};
  PerformanceWatcher sa = new PerformanceWatcher();
  if (debugMode && false)
    PApplet.runSketch(args, sa);

  timeKeeper = new TimeKeeper(sa, true);
  timeKeeper.pw = sa;
}

public class TimeKeeper {
  int t0;
  int tEnd;
  int total = 0;

  IntList times;
  StringList methods;
  int fps = 0;
  PerformanceWatcher pw;

  TimeKeeper(PerformanceWatcher pw, boolean process) {
    if (process) {
      registerMethod("pre", this);
      registerMethod("post", this);
    }
    times = new IntList();
    methods = new StringList();
    this.pw = pw;
  }

  public void pre() {
    times = new IntList();
    methods = new StringList();
    t0 = millis();
  }

  public void post() {
    tEnd = millis();
    total = tEnd-t0;
    fps = (int)frameRate;
    //println("TIMER total:", total + " (" + fps + " fps)");
    for (int i = 0; i < times.size(); i++) {
      //println("\t" + methods.get(i) + ": " + times.get(i) + " (" + (int)((float)times.get(i)/total*100) + "%)");
    }
    this.clone();
    if (count++ % 10 == 0) {
      pw.drawIt(this.clone());
    }
    //pw.drawNum(total);
  }
  int count = 0;

  public void add(int time, String methodName) {
    times.append(time);
    methods.append(methodName);
  }

  public TimeKeeper clone() {
    TimeKeeper clone = new TimeKeeper(this.pw, false);
    for (int i = 0; i < times.size(); i++) {

      clone.times.append(times.get(i));
      clone.methods.append(methods.get(i));
    }
    clone.total = total;
    clone.fps = fps;
    return clone;
  }
}



public class PerformanceWatcher extends PApplet {

  int rectWidth;
  PGraphics x;

  public void setup() {
    surface.setSize(200, 200);
    x = createGraphics(width, height) ;
    surface.setAlwaysOnTop(true);
    surface.setLocation(displayWidth-width -20, displayHeight-height -50);
    x.colorMode(HSB);
    rectWidth = width/2;
    noLoop();
  }
  public void draw() {
    this.image(x, 0, 0);
  }

  int count = 0;
  public void drawIt(TimeKeeper a) {
    if (x!=null) {


      count = 0;
      x.beginDraw();
      x.background(255);
      int number = a.times.size();
      float usedSpace = 0;

      int clrRunner = 0;

      for (int i = 0; i < number; i++) {
        color c = color(clrRunner, 120, 120);
        float percentage;
        if (a.methods.get(i).equals("draw")) {
          percentage= 1-((float)a.times.get(i)/(float)a.total);
        } else {
          percentage = ((float)a.times.get(i)/(float)a.total);
        }

        x.fill(c);
        x.rect(0, usedSpace, rectWidth, percentage*height);

        x.textAlign(LEFT, TOP);
        x.text(a.methods.get(i), rectWidth+10, i*20);
        x.textAlign(RIGHT, TOP);
        x.text((int)(100*percentage) + "%", width, i*20);

        x.fill(0);
        x.textAlign(CENTER, TOP);
        x.text(a.methods.get(i), rectWidth/2, usedSpace);
        usedSpace += percentage * height;
        clrRunner = (clrRunner + 50)%255;
      }
      x.text(a.total + " (" + a.fps + " fps)", width-40, height-20);
      x.endDraw();
      redraw();
    }
  }


  public void drawNum(int a) {
    if (x !=null) {
      x.beginDraw();
      this.redraw();
      println(2);
      x.background(0);
      x.text(a, 50, 50);
      x.endDraw();
      this.redraw();
    }
  }
}
