import pGUI.classes.*; //<>//
import pGUI.core.*;
import java.awt.Rectangle; 



Settings setting;


void setup() {
  //size(10,10);
  fullScreen();




  frameRate(60);                         //will be set later, because high fps will be needed for starting animation

  camera = new CamContainer();           //no initializing yet, still a list of available cameras is being fetched in another thread already
  camera.selectCamera(camNumber);        //select camera index from list

  flexSensor = new FlexSensor(this);
  dir = new Direction();
  notifier = new Notifier(250, 60);


  //set up the gui and window properties
  setupVisuals();
  setting = new Settings();


  registerMethod("mouseEvent", this);
  registerMethod("keyEvent", this);

  //initialize the tiny "taskmanager"
  initTimeKeeper();
  //calibBGColor = color(250, 0, 250);
  colorMode(HSB, width, 100, height);
  calibBGColor = color(714, 150, 51);
}


//change background color when mouse is moved - for testing purposes
void mouseMoved() {
  //colorMode(HSB, width, 100, height);
  //calibBGColor = color(mouseX, 100, mouseY);
  //println(mouseX, mouseY);
}



/*__________________D R A W____________________________*/

void draw() {
  int t0 = millis();
  if (camera.isInitialized()) { //cam needs to be initialized, aka get the first image, so image dimensions can be passed to featureDetection
    if (tracker == null) {


      tracker = new EyeTracker(this, camera);
      textAlign(LEFT);
      frameRate(generalFrameRate);

      //
    } else {

      if (tracker.calibrating) {
        /* 
         * Calibrate Eyetracker
         */
        tracker.calibrate();     //run through the calibration process

        //
      } else {

        //MAIN STUFF, RUNNING MOST OF THE TIME: 



        if (frame.getVisible() == false) {    //show GUI
          frame.setVisible(true);
          textArea.setEnabled(true);
        }


        background(background);
        fill(color(200, 170, 170, 200));
        noStroke();
        rect(dir.current % 3 * (width/3), dir.current / 3 *(height/3), width/3, height/3);
        ECScreenConstants.drawScreen();       //draw background instruction screens

        if (darkMode) filter(INVERT);         //invert image for dark mode


        if (!useMouse) {
          tracker.track2();                   //main function, runs most of the time, finds out iris position
          dir.current = tracker.evaluate2();  //get direction out of iris position
        } else {
          analyzeMouseMovement();
        }


        analyzeEyeData();                     //take direction, compare it to previous directions and send it to UI-Interaction-Interpreters

        analyzeArduinoData();                 //check if flexsensor has signal, if so send to UI-Interaction-Interpreters





        //diplay the tolerances in the 4 direction - for debugging and testing
        if (debugMode) {
          text(tracker.tU, 1 /2. * (width-100)+50, 0 / 2. *(height-100)+50);
          text(tracker.tD, 1 /2. * (width-100)+50, 2 / 2. *(height-100)+50);
          text(tracker.tR, 2 /2. * (width-100)+50, 1 / 2. *(height-100)+50);
          text(tracker.tL, 0 /2. * (width-100)+50, 1 / 2. *(height-100)+50);
        }
      }
    }
  } else { //before camera is initialized show this animation
    loadingAnimation();
  }
  timeKeeper.add(millis()-t0, "draw");  //just keep track of how much time is needed for the draw method
}









/* Animation that is shown while cameras are being searched  */

PImage loadingImage;
int loadSize;         //size for loading png
int loadHeight;
int rotateAngle = 0;  //angle for loading animation

void loadingAnimation() {

  if (loadingImage == null) {
    loadingImage = loadImage("loading.png");
    loadSize = width/8;
  }

  frameRate(60);
  background(calibBGColor);
  camera.initCam(this);      //checks if camera list is ready and then starts initializing
  textAlign(CENTER);
  textSize(40);
  text("Searching for cameras...", width/2, height/5);

  translate(width/2, height/2);
  rotate(rotateAngle*TWO_PI/360);
  translate(-loadSize/2, -loadSize/2);
  image(loadingImage, 0, 0, loadSize, loadSize);

  rotateAngle += 2;
}
